# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Handcrafted features for linear models."""

import dataclasses

import librosa
from ml_collections import config_dict
import numpy as np
from perch_hoplite.zoo import zoo_interface


@dataclasses.dataclass
class HandcraftedFeaturesModel(zoo_interface.EmbeddingModel):
  """Wrapper for simple feature extraction.

  Parameters:
    window_size_s: The window size for framing the audio.
    hop_size_s: The hop size for framing the audio.
    features_config: The configuration for the features. Used as arguments to
      librosa.feature.melspectrogram or librosa.feature.mfcc, depending on
      `use_mfccs`.
    use_mfccs: Whether to use MFCCs or melspectrogram.
    aggregation: How to aggregate the features.
  """

  window_size_s: float
  hop_size_s: float
  features_config: config_dict.ConfigDict
  use_mfccs: bool = True
  aggregation: str = 'beans'

  @classmethod
  def from_config(
      cls, config: config_dict.ConfigDict
  ) -> 'HandcraftedFeaturesModel':
    return cls(**config)

  @classmethod
  def beans_baseline_config(
      cls, sample_rate=32000, frame_rate=100
  ) -> config_dict.ConfigDict:
    stride = sample_rate // frame_rate
    features_config = config_dict.ConfigDict({
        'sample_rate': sample_rate,
        'n_mfcc': 20,
        'hop_length': stride,
        'win_length': 2 * stride,
        'freq_range': (60.0, sample_rate / 2.0),
        'power': 2.0,
    })
    return config_dict.ConfigDict({
        'sample_rate': sample_rate,
        'features_config': features_config,
        'use_mfccs': True,
        'window_size_s': 1.0,
        'hop_size_s': 1.0,
        'aggregation': 'beans',
    })

  @classmethod
  def beans_baseline(
      cls, sample_rate=32000, frame_rate=100
  ) -> 'HandcraftedFeaturesModel':
    config = cls.beans_baseline_config(sample_rate, frame_rate)
    # pylint: disable=unexpected-keyword-arg
    return HandcraftedFeaturesModel.from_config(config)

  def compute_features(self, audio_array: np.ndarray) -> np.ndarray:
    framed_audio = self.frame_audio(
        audio_array, self.window_size_s, self.hop_size_s
    )
    specs = []
    for frame in framed_audio:
      if self.use_mfccs:
        features = librosa.feature.mfcc(
            y=frame,
            sr=self.sample_rate,
            hop_length=self.features_config.hop_length,
            win_length=self.features_config.win_length,
            center=True,
            norm='ortho',
            dct_type=2,
            n_mfcc=self.features_config.n_mfcc,
        )
      else:
        features = librosa.feature.melspectrogram(
            y=frame,
            sr=self.sample_rate,
            hop_length=self.features_config.hop_length,
            win_length=self.features_config.win_length,
            center=True,
            n_mels=self.features_config.n_mels,
            power=self.features_config.power,
        )
      specs.append(features)
    return np.stack(specs, axis=0)

  def embed(self, audio_array: np.ndarray) -> zoo_interface.InferenceOutputs:
    # Melspecs will have shape [melspec_channels, frames]
    features = self.compute_features(audio_array)
    print(f'features: {features.shape}')
    if self.aggregation == 'beans':
      features = np.concatenate(
          [
              features.mean(axis=-1),
              features.std(axis=-1),
              features.min(axis=-1),
              features.max(axis=-1),
          ],
          axis=-1,
      )
    else:
      raise ValueError(f'unrecognized aggregation: {self.aggregation}')
    # Add a trivial channels dimension.
    features = features[:, np.newaxis, :]
    return zoo_interface.InferenceOutputs(features, None, None)

  def batch_embed(
      self, audio_batch: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    return zoo_interface.batch_embed_from_embed_fn(self.embed, audio_batch)



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementations of inference interfaces for applying trained models."""

import dataclasses

from ml_collections import config_dict
import numpy as np
from perch_hoplite.taxonomy import namespace_db
from perch_hoplite.zoo import zoo_interface


@dataclasses.dataclass
class PlaceholderModel(zoo_interface.EmbeddingModel):
  """Test implementation of the EmbeddingModel zoo_interface."""

  embedding_size: int = 128
  make_embeddings: bool = True
  make_logits: bool = True
  make_separated_audio: bool = True
  make_frontend: bool = True
  do_frame_audio: bool = False
  window_size_s: float = 1.0
  hop_size_s: float = 1.0
  frontend_size: tuple[int, int] = (32, 32)

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict) -> 'PlaceholderModel':
    return cls(**config)

  def __post_init__(self):
    db = namespace_db.load_db()
    self.class_list = db.class_lists['caples']

  def embed(self, audio_array: np.ndarray) -> zoo_interface.InferenceOutputs:
    outputs = {}
    if self.do_frame_audio:
      audio_array = self.frame_audio(
          audio_array, self.window_size_s, self.hop_size_s
      )
    time_size = audio_array.shape[0] // int(
        self.window_size_s * self.sample_rate
    )
    if self.make_embeddings:
      outputs['embeddings'] = np.zeros(
          [time_size, 1, self.embedding_size], np.float32
      )
    if self.make_frontend:
      outputs['frontend'] = np.zeros(
          [time_size, self.frontend_size[0], self.frontend_size[1]], np.float32
      )
    if self.make_logits:
      outputs['logits'] = {
          'label': np.zeros(
              [time_size, len(self.class_list.classes)], np.float32
          ),
          'other_label': np.ones(
              [time_size, len(self.class_list.classes)], np.float32
          ),
      }
    if self.make_separated_audio:
      outputs['separated_audio'] = np.zeros(
          [2, audio_array.shape[-1]], np.float32
      )
    return zoo_interface.InferenceOutputs(**outputs)

  def batch_embed(
      self, audio_batch: np.ndarray
  ) -> zoo_interface.InferenceOutputs:
    return zoo_interface.batch_embed_from_embed_fn(self.embed, audio_batch)



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Interface for models producing embeddings."""

import dataclasses
import json
from typing import Any, Callable, Dict

from absl import logging
from etils import epath
import librosa
from ml_collections import config_dict
import numpy as np
from perch_hoplite.taxonomy import namespace
import tensorflow as tf

LogitType = Dict[str, np.ndarray]

NULL_LOGIT = -20.0
POOLING_METHODS = ['first', 'mean', 'max', 'mid', 'flatten', 'squeeze']


@dataclasses.dataclass
class InferenceOutputs:
  """Wrapper class for outputs from an inference model.

  Attributes:
    embeddings: Embeddings array with shape [Frames, Channels, Features].
    logits: Dictionary mapping a class list L's name to an array of logits. The
      logits array has shape [Frames, L.size] or [Frames, Channels, L.size].
    separated_audio: Separated audio channels with shape [Channels, Samples].
    batched: If True, each output has an additonal batch dimension.
    frontend: Model frontend output, such as melspectrogram.
  """

  embeddings: np.ndarray | None = None
  logits: LogitType | None = None
  separated_audio: np.ndarray | None = None
  batched: bool = False
  frontend: np.ndarray | None = None

  def __post_init__(self):
    # In some scenarios, we may be passed TF EagerTensors. We dereference these
    # to numpy arrays for broad compatibility.
    if hasattr(self.embeddings, 'numpy'):
      self.embeddings = self.embeddings.numpy()
    if self.logits is not None:
      for k, v in self.logits.items():
        if hasattr(v, 'numpy'):
          self.logits[k] = v.numpy()
    if hasattr(self.separated_audio, 'numpy'):
      self.separated_audio = self.separated_audio.numpy()

  def pooled_embeddings(
      self, time_pooling: str, channel_pooling: str = ''
  ) -> np.ndarray:
    """Reduce embeddings over the time and/or channel axis."""
    # Shape is either [B, F, C, D] or [F, C, D], so the time axis is -3.
    outputs = pool_axis(self.embeddings, -3, time_pooling)
    outputs = pool_axis(outputs, -2, channel_pooling)
    return outputs


EmbedFnType = Callable[[np.ndarray], InferenceOutputs]


@dataclasses.dataclass
class EmbeddingModel:
  """Wrapper for a model which produces audio embeddings.

  It is encouraged to implement either the `embed` or `batch_embed` function
  and use a convenience method (`batch_embed_from_embed_fn` or
  `embed_from_batch_embed_fn`) to get the other. It is preferable to implement
  `batch_embed` so long as the model accepts batch input, as batch input
  inference can be much faster.

  Attributes:
    sample_rate: Sample rate in hz.
  """

  sample_rate: int

  @classmethod
  def from_config(
      cls, model_config: config_dict.ConfigDict
  ) -> 'EmbeddingModel':
    """Load the model from a configuration dict."""
    raise NotImplementedError

  def embed(self, audio_array: np.ndarray) -> InferenceOutputs:
    """Create InferenceOutputs from an audio array.

    Args:
      audio_array: An array with shape [Time] containing unit-scaled audio.

    Returns:
      An InferenceOutputs object.
    """
    raise NotImplementedError

  def batch_embed(self, audio_batch: np.ndarray) -> InferenceOutputs:
    """Create InferenceOutputs from a batch of audio arrays.

    Args:
      audio_batch: An array with shape [Time] containing unit-scaled audio.

    Returns:
      An InferenceOutputs object.
    """
    raise NotImplementedError

  def frame_audio(
      self,
      audio_array: np.ndarray,
      window_size_s: float | None,
      hop_size_s: float,
  ) -> np.ndarray:
    """Helper function for framing audio for inference along the last axis."""
    if window_size_s is None or window_size_s < 0:
      return np.expand_dims(audio_array, axis=-2)
    frame_length = int(window_size_s * self.sample_rate)
    hop_length = int(hop_size_s * self.sample_rate)
    if audio_array.shape[-1] < frame_length:
      audio_array = librosa.util.pad_center(
          audio_array, size=frame_length, axis=-1
      )
    # Librosa frames as [..., frame_length, frames], so we need a transpose.
    framed_audio = librosa.util.frame(
        audio_array,
        frame_length=frame_length,
        hop_length=hop_length,
        axis=-1,
    ).swapaxes(-1, -2)
    return framed_audio

  def normalize_audio(
      self,
      framed_audio: np.ndarray,
      target_peak: float | None,
  ) -> np.ndarray:
    """Normalizes audio with shape [..., T] to match the target_peak value."""
    if target_peak is None:
      return framed_audio
    framed_audio = framed_audio.copy()
    framed_audio -= np.mean(framed_audio, axis=-1, keepdims=True)
    peak_norm = np.max(np.abs(framed_audio), axis=-1, keepdims=True)
    framed_audio = np.divide(framed_audio, peak_norm, where=(peak_norm > 0.0))
    framed_audio = framed_audio * target_peak
    return framed_audio


@dataclasses.dataclass
class LogitsOutputHead:
  """A TensorFlow model which classifies embeddings.

  Attributes:
    model_path: Path to saved model.
    logits_key: Name of this output head.
    logits_model: Callable model converting embeddings of shape [B,
      embedding_width] to [B, num_classes].
    class_list: ClassList specifying the ordered classes.
    channel_pooling: Pooling to apply to channel dimension of logits. Specify an
      empty string to apply no pooling.
  """

  model_path: str
  logits_key: str
  logits_model: Any
  class_list: namespace.ClassList
  channel_pooling: str = 'max'

  @classmethod
  def from_config_file(cls, model_path: str, filename='logits_config.json'):
    config_filepath = epath.Path(model_path) / filename
    with (config_filepath).open() as f:
      logits_config = config_dict.ConfigDict(json.loads(f.read()))
    logits_config.model_path = model_path
    return cls.from_config(logits_config)

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict):
    logits_model = tf.saved_model.load(config.model_path)
    model_path = epath.Path(config.model_path)
    with (model_path / 'class_list.csv').open('r') as f:
      class_list = namespace.ClassList.from_csv(f)
    return cls(
        logits_model=logits_model,
        class_list=class_list,
        **config,
    )

  def __call__(self, embeddings: np.ndarray) -> InferenceOutputs:
    """Apply the wrapped logits_model to embeddings with shape [B, D]."""
    if callable(self.logits_model):
      logits = self.logits_model(embeddings)
    elif hasattr(self.logits_model, 'signatures'):
      # TODO(tomdenton): Figure out why the Keras saved model isn't callable.
      flat_logits = self.logits_model.signatures['serving_default'](
          inputs=embeddings
      )
      logits = flat_logits['output_0']
      if hasattr(logits, 'numpy'):
        logits = logits.numpy()
    else:
      raise ValueError('could not figure out how to call wrapped model.')
    return logits

  def save_model(self, output_path: str, embeddings_path: str):
    """Write a SavedModel and metadata to disk."""
    # Write the model.
    tf.saved_model.save(self.logits_model, output_path)
    output_path = epath.Path(output_path)
    # Write a config file.
    config_data = dataclasses.asdict(self)
    for k in ['logits_model', 'class_list']:
      # These are loaded automatically.
      config_data.pop(k)
    with (output_path / 'logits_config.json').open('w') as f:
      json.dump(config_data, f)
    # Copy the embeddings_config if provided
    if embeddings_path:
      (epath.Path(embeddings_path) / 'config.json').copy(
          output_path / 'embeddings_config.json', overwrite=True
      )
    # Write the class list.
    with (output_path / 'class_list.csv').open('w') as f:
      f.write(self.class_list.to_csv())

  def add_logits(self, model_outputs: InferenceOutputs, keep_original: bool):
    """Update the model_outputs to include logits from this output head."""
    embeddings = model_outputs.embeddings
    if embeddings is None:
      logging.warning('No embeddings found in model outputs.')
      return model_outputs
    flat_embeddings = np.reshape(embeddings, [-1, embeddings.shape[-1]])
    flat_logits = self(flat_embeddings)
    logits_shape = np.concatenate(
        [np.shape(embeddings)[:-1], np.shape(flat_logits)[-1:]], axis=0
    )
    logits = np.reshape(flat_logits, logits_shape)
    # Embeddings have shape [B, T, C, D] or [T, C, D], so our logits also
    # have a channel dimension.
    # Output logits should have shape [B, T, D] or [T, D], so we reduce the
    # channel axis as specified by the user.
    # The default is 'max' which is reasonable for separated audio and
    # is equivalent to 'squeeze' for the single-channel case.
    logits = pool_axis(logits, -2, self.channel_pooling)
    new_outputs = InferenceOutputs(
        embeddings=model_outputs.embeddings,
        logits=model_outputs.logits,
        separated_audio=model_outputs.separated_audio,
        batched=model_outputs.batched,
    )
    if new_outputs.logits is None or not keep_original:
      new_outputs.logits = {}
    new_outputs.logits[self.logits_key] = logits
    return new_outputs


def embed_from_batch_embed_fn(
    embed_fn: EmbedFnType, audio_array: np.ndarray
) -> InferenceOutputs:
  """Embed a single example using a batch_embed_fn."""
  audio_batch = audio_array[np.newaxis, :]
  outputs = embed_fn(audio_batch)

  unbatched_outputs = {}
  for k in ['embeddings', 'separated_audio', 'frontend']:
    if getattr(outputs, k) is not None:
      unbatched_outputs[k] = getattr(outputs, k)[0]
    else:
      unbatched_outputs[k] = None

  if outputs.logits is not None:
    logits = {}
    for k, v in outputs.logits.items():
      logits[k] = v[0]
  else:
    logits = None

  return InferenceOutputs(
      logits=logits,
      batched=False,
      **unbatched_outputs,
  )


def batch_embed_from_embed_fn(
    embed_fn: EmbedFnType, audio_batch: np.ndarray
) -> InferenceOutputs:
  """Embed a batch of audio using a single-example embed_fn."""
  outputs = []
  for audio in audio_batch:
    outputs.append(embed_fn(audio))

  batched_outputs = {}
  for k in ['embeddings', 'separated_audio', 'frontend']:
    if getattr(outputs[0], k) is not None:
      batched_outputs[k] = np.stack([getattr(x, k) for x in outputs], axis=0)
    else:
      batched_outputs[k] = None

  if outputs[0].logits is not None:
    batched_logits = {}
    for logit_key in outputs[0].logits:
      batched_logits[logit_key] = np.stack(
          [x.logits[logit_key] for x in outputs], axis=0
      )
  else:
    batched_logits = None

  return InferenceOutputs(
      logits=batched_logits,
      batched=True,
      **batched_outputs,
  )


def pool_axis(ar: np.ndarray, axis: int, pooling: str) -> np.ndarray:
  """Apply the specified pooling along the target axis."""
  if pooling == 'first':
    outputs = ar.take(0, axis=axis)
  elif pooling == 'squeeze':
    # Like 'first' but throws an exception if more than one time step.
    outputs = ar.squeeze(axis=axis)
  elif pooling == 'mean':
    outputs = ar.mean(axis=axis)
  elif pooling == 'max':
    outputs = ar.max(axis=axis)
  elif pooling == 'mid':
    midpoint_index = ar.shape[axis] // 2
    outputs = ar.take(midpoint_index, axis=axis)
  elif pooling == 'flatten':
    # Flatten the target axis dimension into the last dimension.
    outputs = ar.swapaxes(axis, -2)
    new_shape = outputs.shape[:-2] + (outputs.shape[-1] * outputs.shape[-2],)
    outputs = outputs.reshape(new_shape)
  elif not pooling:
    outputs = ar
  else:
    raise ValueError(f'Unrecognized pooling method {pooling}.')
  return outputs



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Audio IO utilities."""

import concurrent
import itertools
import logging
import os
import tempfile
from typing import Callable, Generator, Sequence
import warnings

from etils import epath
import librosa
import numpy as np
from perch_hoplite import path_utils
import requests
import soundfile


def load_audio(
    path: epath.PathLike,
    target_sample_rate: int,
    dtype: str = 'float32',
    **kwargs,
) -> np.ndarray:
  """Load a general audio resource."""
  path = os.fspath(path)
  if path.startswith('xc'):
    return load_xc_audio(path, target_sample_rate, dtype=dtype)
  elif path.startswith('http'):
    return load_url_audio(path, target_sample_rate, dtype=dtype)
  else:
    return load_audio_file(path, target_sample_rate, dtype=dtype, **kwargs)


def load_audio_file(
    filepath: str | epath.Path,
    target_sample_rate: int,
    resampling_type: str = 'polyphase',
    dtype: str = 'float32',
) -> np.ndarray:
  """Read an audio file, and resample it using librosa."""
  filepath = epath.Path(filepath)
  if target_sample_rate <= 0:
    # Use the native sample rate.
    target_sample_rate = None
  if expect_soundfile_compatibility(filepath):
    with filepath.open('rb') as f:
      sf = soundfile.SoundFile(file=f)
      audio = sf.read()
      if target_sample_rate is not None:
        audio = librosa.resample(
            y=audio,
            orig_sr=sf.samplerate,
            target_sr=target_sample_rate,
            res_type=resampling_type,
        )
      return audio.astype(dtype)

  # Handle other audio formats.
  # Because librosa passes file handles to soundfile, we need to copy the file
  # to a temporary file before passing it to librosa.
  extension = epath.Path(filepath).suffix.lower()
  with tempfile.NamedTemporaryFile(
      mode='w+b', suffix=extension, delete=False
  ) as f:
    with filepath.open('rb') as sf:
      f.write(sf.read())
  # librosa outputs lots of warnings which we can safely ignore when
  # processing all Xeno-Canto files and PySoundFile is unavailable.
  with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    audio, _ = librosa.load(
        f.name,
        sr=target_sample_rate,
        res_type=resampling_type,
    )
  os.unlink(f.name)
  return audio.astype(dtype)


def load_audio_window_soundfile(
    filepath: str,
    offset_s: float,
    sample_rate: int,
    window_size_s: float,
    dtype: str = 'float32',
) -> np.ndarray:
  """Load an audio window using Soundfile.

  Args:
    filepath: Path to audio file.
    offset_s: Read offset within the file.
    sample_rate: Sample rate for returned audio.
    window_size_s: Length of audio to read. Reads all if <0.
    dtype: Audio dtype.

  Returns:
    Numpy array of loaded audio.
  """
  with epath.Path(filepath).open('rb') as f:
    sf = soundfile.SoundFile(f)
    if offset_s > 0:
      offset = int(np.float32(offset_s) * sf.samplerate)
      sf.seek(offset)
    if window_size_s < 0:
      a = sf.read()
    else:
      window_size = int(window_size_s * sf.samplerate)
      a = sf.read(window_size)
  if len(a.shape) == 2:
    # Downstream ops expect mono audio, so reduce to mono.
    a = a[:, 0]
  if sample_rate > 0:
    a = librosa.resample(
        y=a, orig_sr=sf.samplerate, target_sr=sample_rate, res_type='polyphase'
    )
  return a.astype(dtype)


def load_audio_window(
    filepath: str,
    offset_s: float,
    sample_rate: int,
    window_size_s: float,
    dtype: str = 'float32',
) -> np.ndarray:
  """Load a slice of audio from a file, hopefully efficiently."""

  if expect_soundfile_compatibility(filepath):
    try:
      return load_audio_window_soundfile(
          filepath, offset_s, sample_rate, window_size_s, dtype
      )
    except soundfile.LibsndfileError:
      logging.info('Failed to load audio with libsndfile: %s', filepath)

  # Try librosa.
  try:
    duration = window_size_s if window_size_s > 0 else None
    audio, _ = librosa.load(
        filepath, sr=sample_rate, offset=offset_s, duration=duration
    )
    return audio.astype(dtype)
  except Exception as exc:  # pylint: disable=broad-exception-caught
    logging.error('Failed to load audio with librosa (%s) : %s.', filepath, exc)

  # This fail-over is much slower but more reliable; the entire audio file
  # is loaded (and possibly resampled) and then we extract the target audio.
  audio = load_audio(filepath, sample_rate)
  offset = int(offset_s * sample_rate)
  window_size = int(window_size_s * sample_rate)
  return audio[offset : offset + window_size].astype(dtype)


def multi_load_audio_window(
    filepaths: Sequence[str],
    offsets: Sequence[float] | None,
    audio_loader: Callable[[str, float], np.ndarray],
    max_workers: int = 5,
    buffer_size: int = -1,
) -> Generator[np.ndarray, None, None]:
  """Generator for loading audio windows in parallel.

  Note that audio is returned in the same order as the filepaths.
  Also, this ultimately relies on soundfile, which can be buggy in some cases.

  Caution: Because this generator uses an Executor, it can continue holding
  resources while not being used. If you are using this in a notebook, you
  should use this in a 'nameless' context, like:
  ```
  for audio in multi_load_audio_window(...):
    ...
  ```
  or in a try/finally block:
  ```
  audio_iterator = multi_load_audio_window(...)
  try:
    for audio in audio_iterator:
      ...
  finally:
    del(audio_iterator)
  ```
  Otherwise, the generator will continue to hold resources until the notebook
  is closed.

  Args:
    filepaths: Paths to audio to load.
    offsets: Read offset in seconds for each file, or None if no offsets are
      needed.
    audio_loader: Function to load audio given a filepath and offset.
    max_workers: Number of threads to allocate.
    buffer_size: Max number of audio windows to queue up. Defaults to 10x the
      number of workers.

  Yields:
    Loaded audio windows.
  """
  if buffer_size == -1:
    buffer_size = 10 * max_workers
  if offsets is None:
    offsets = [0.0 for _ in filepaths]

  # TODO(tomdenton): Use itertools.batched in Python 3.12+
  def batched(iterable, n):
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
      yield batch

  task_iterator = zip(filepaths, offsets)
  batched_iterator = batched(task_iterator, buffer_size)
  mapping = lambda x: audio_loader(x[0], x[1])

  executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
  try:
    yield from itertools.chain.from_iterable(
        executor.map(mapping, batch) for batch in batched_iterator
    )
  finally:
    executor.shutdown(wait=False, cancel_futures=True)


def load_xc_audio(
    xc_id: str, sample_rate: int, dtype: str = 'float32'
) -> np.ndarray:
  """Load audio from Xeno-Canto given an ID like 'xc12345'."""
  if not xc_id.startswith('xc'):
    raise ValueError(f'XenoCanto id {xc_id} does not start with "xc".')
  xc_id = xc_id[2:]
  try:
    int(xc_id)
  except ValueError as exc:
    raise ValueError(f'XenoCanto id xc{xc_id} is not an integer.') from exc
  session = requests.Session()
  session.mount(
      'https://',
      requests.adapters.HTTPAdapter(
          max_retries=requests.adapters.Retry(total=5, backoff_factor=0.1)
      ),
  )
  url = f'https://xeno-canto.org/{xc_id}/download'
  try:
    data = session.get(url=url).content
  except requests.exceptions.RequestException as e:
    raise requests.exceptions.RequestException(
        f'Failed to load audio from Xeno-Canto {xc_id}'
    ) from e
  with tempfile.NamedTemporaryFile(suffix='.mp3', mode='wb', delete=False) as f:
    f.write(data)
    f.flush()
  audio = load_audio_file(f.name, target_sample_rate=sample_rate)
  os.unlink(f.name)
  return audio.astype(dtype)


def load_url_audio(
    url: str, sample_rate: int, dtype: str = 'float32'
) -> np.ndarray:
  """Load audio from a URL."""
  data = requests.get(url).content
  with tempfile.NamedTemporaryFile(mode='wb', delete=False) as f:
    f.write(data)
    f.flush()
  audio = load_audio_file(f.name, target_sample_rate=sample_rate)
  os.unlink(f.name)
  return audio.astype(dtype)


def get_file_length_s_and_sample_rate(filepath: str) -> tuple[float, int]:
  """As it says on the tin, or (-1, -1) if unparseable."""
  if expect_soundfile_compatibility(filepath):
    try:
      with epath.Path(filepath).open('rb') as f:
        sf = soundfile.SoundFile(f)
        file_length_s = sf.frames / sf.samplerate
        return file_length_s, sf.samplerate
    except Exception as exc:  # pylint: disable=broad-exception-caught
      logging.error('Failed to parse audio file (%s) : %s.', filepath, exc)

  # Try librosa.
  try:
    sr = librosa.get_samplerate(filepath)
    file_length_s = librosa.get_duration(filename=filepath)
    return file_length_s, sr
  except Exception as exc:  # pylint: disable=broad-exception-caught
    logging.error('Failed to parse audio file (%s) : %s.', filepath, exc)

  return -1, -1


def expect_soundfile_compatibility(filepath: str | epath.PathLike) -> bool:
  """Returns True if soundfile can be used to load the audio file."""
  extension = epath.Path(filepath).suffix.lower()
  if extension in ('.wav', '.flac', '.ogg', '.opus'):
    try:
      with epath.Path(filepath).open('rb') as f:
        soundfile.SoundFile(f)
        return True
    except soundfile.LibsndfileError:
      pass
  return False



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for handling namespaces of classes."""
from __future__ import annotations

import csv
import dataclasses
import io
from typing import Iterable

import numpy as np

UNKNOWN_LABEL = "unknown"


@dataclasses.dataclass
class Namespace:
  """A namespace is simply a set of labels.

  Note that unknown labels cannot be in a namespace.

  Attributes:
    classes: A frozenset of labels.
  """

  classes: frozenset[str]

  def __post_init__(self):
    if UNKNOWN_LABEL in self.classes:
      raise ValueError("unknown class")

  def __str__(self):
    example_class = sorted(list(self.classes))[0]
    return (
        f"Namespace with {len(self.classes)} classes, "
        f"including '{example_class}'"
    )


@dataclasses.dataclass
class Mapping:
  """A mapping maps labels from one namespace to labels in another.

  Note that this is a n:1 mapping, i.e., multiple labels in the source namespace
  can map to the same label in the target namespace.

  The source and target namespace are referred to by their name. This name must
  be resolved using the taxonomy database.

  Note that labels (other than unknown) cannot be mapped to unknown. Instead,
  these labels should be simply excluded from the mapping. The end-user is
  responsible for deciding whether to map missing keys to unknown or whether to
  raise an error, e.g., by using:

    mapping.mapped_pairs.get(source_label, namespace.UNKNOWN_LABEL)

  Attributes:
    source_namespace: The name of the source namespace.
    target_namespace: The name of the target namespace.
    mapped_pairs: The mapping from labels in the source namespace to labels in
      the target namespace.
  """

  source_namespace: str
  target_namespace: str
  mapped_pairs: dict[str, str]

  def __post_init__(self):
    if UNKNOWN_LABEL in self.mapped_pairs.values():
      raise ValueError("unknown target class")

  def __str__(self):
    return f"Mapping from {self.source_namespace} to {self.target_namespace}."

  def reverse(self) -> "Mapping":
    """Returns a mapping with the source and target namespaces swapped."""
    if len(set(self.mapped_pairs.values())) != len(self.mapped_pairs.values()):
      raise ValueError("Mapping is not injective, cannot reverse.")
    return Mapping(
        source_namespace=self.target_namespace,
        target_namespace=self.source_namespace,
        mapped_pairs={v: k for k, v in self.mapped_pairs.items()},
    )


@dataclasses.dataclass
class ClassList:
  """A list of labels.

  A class list is a list of labels in a particular order, e.g., to reflect the
  output of a model.

  Class lists can contain the unknown label. All other labels must belong to a
  namespace.

  Class lists cannot contain duplicate entries.

  Attributes:
    namespace: The name of the namespace these class labels belong to.
    classes: The list of classes.
  """

  namespace: str
  classes: tuple[str, ...]

  def __post_init__(self):
    if len(set(self.classes)) != len(self.classes):
      raise ValueError("duplicate entries in class list")

  def __str__(self):
    example_class = sorted(list(self.classes))[0]
    return (
        f"ClassList in namespace {self.namespace} with {len(self.classes)}"
        f" classes, including '{example_class}'."
    )

  @classmethod
  def from_csv(cls, csv_data: Iterable[str]) -> "ClassList":
    """Parse a class list from a CSV file.

    The file must contain the namespace in the first column of the first row.
    The first column of the remaining rows are assumed to contain the classes.

    Args:
      csv_data: Any iterable which can be passed on to `csv.reader`.

    Returns:
      The parsed class list.
    """
    reader = csv.reader(csv_data)
    namespace = next(reader)[0]
    classes = tuple(row[0].strip() for row in reader if row)
    return ClassList(namespace, classes)

  def to_csv(self) -> str:
    """Write a class list to a CSV file.

    See `from_csv` for a description of the file format.

    It can be useful to write the class lists to disk so that the model can be
    loaded correctly, even if class lists change. However, note that in this
    case none of the mappings are guaranteed to still work.

    Returns:
      A string containing the namespace and the class labels as rows.
    """
    buffer = io.StringIO(newline="")
    writer = csv.writer(buffer)
    writer.writerow([self.namespace])
    for class_ in self.classes:
      writer.writerow([class_])
    return buffer.getvalue()

  def apply_namespace_mapping(
      self, mapping: Mapping, keep_unknown: bool | None = None
  ) -> ClassList:
    """Apply a namespace mapping to this class list.

    Args:
      mapping: The mapping to apply.
      keep_unknown: How to handle unknowns. If true, then unknown labels in the
        class list are maintained as unknown in the mapped values. If false then
        the unknown value is discarded. The default (`None`) will raise an error
        if an unknown value is in the source classt list.

    Returns:
      A class list which is the result of applying the given mapping to this
      class list.

    Raises:
      KeyError: If a class in not the mapping, or if the class list contains
      an unknown token and `keep_unknown` was not specified.
    """
    if mapping.source_namespace != self.namespace:
      raise ValueError("mapping source namespace does not match class list's")
    mapped_pairs = mapping.mapped_pairs
    if keep_unknown:
      mapped_pairs = mapped_pairs | {UNKNOWN_LABEL: UNKNOWN_LABEL}
    return ClassList(
        mapping.target_namespace,
        tuple(
            dict.fromkeys(
                mapped_pairs[class_]
                for class_ in self.classes
                if class_ != UNKNOWN_LABEL or keep_unknown in (True, None)
            )
        ),
    )

  def get_class_map_matrix(
      self,
      target_class_list: ClassList,
      mapping: Mapping | None = None,
  ) -> tuple[np.ndarray, np.ndarray]:
    """Construct a binary matrix for mapping to another class list.

    Args:
      target_class_list: Class list to map into.
      mapping: Namespace mapping, required if the source and target are in
        different namespaces.

    Returns:
      A binary matrix mapping self to target_class_list and an indicator vector
      for the image of the mapping.
    """
    if self.namespace != target_class_list.namespace and mapping is None:
      raise ValueError(
          "If source and target classes are from different namespaces, a"
          " namespace mapping must be provided."
      )
    elif self.namespace == target_class_list.namespace and mapping is not None:
      raise ValueError(
          "If source and target classes are the same, no mapping should be"
          " provided."
      )
    matrix = np.zeros([len(self.classes), len(target_class_list.classes)])

    target_idxs = {k: i for i, k in enumerate(target_class_list.classes)}
    for i, class_ in enumerate(self.classes):
      if mapping is not None:
        class_ = mapping.mapped_pairs[class_]
      if class_ in target_idxs:
        j = target_idxs[class_]
        matrix = matrix.at[i, j].set(1)
    return matrix, np.any(matrix, axis=0)



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for embedding audio."""

import os
import shutil
import tempfile

from ml_collections import config_dict
from perch_hoplite.agile import embed
from perch_hoplite.agile import source_info
from perch_hoplite.agile.tests import test_utils
from perch_hoplite.db import db_loader

from absl.testing import absltest
from absl.testing import parameterized


class EmbedTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()
    self.db_path = os.path.join(self.tempdir, 'test_db')

  def tearDown(self):
    super().tearDown()
    shutil.rmtree(self.tempdir)

  @parameterized.named_parameters(
      ('in_mem', 'in_mem'),
      ('sqlite_usearch', 'sqlite_usearch'),
  )
  def test_embed_worker(self, db_key):
    classes = ['pos', 'neg']
    filenames = ['foo', 'bar', 'baz']
    test_utils.make_wav_files(self.tempdir, classes, filenames, file_len_s=6.0)

    # Create metadata files.
    with open(
        os.path.join(self.tempdir, 'hoplite_metadata_description.csv'), 'w'
    ) as f:
      f.write(
          'field_name,metadata_level,type,description\n'
          'deployment,deployment,str,Deployment identifier.\n'
          'habitat,deployment,str,Habitat type.\n'
          'recording,recording,str,Recording identifier.\n'
          'mic_type,recording,str,Microphone type.\n'
      )
    with open(
        os.path.join(self.tempdir, 'hoplite_deployments_metadata.csv'), 'w'
    ) as f:
      f.write('deployment,habitat\npos,forest\nneg,grassland\n')
    with open(
        os.path.join(self.tempdir, 'hoplite_recordings_metadata.csv'), 'w'
    ) as f:
      f.write(
          'recording,mic_type\n'
          'pos/foo.wav,MicA\n'
          'pos/bar.wav,MicA\n'
          'pos/baz.wav,MicB\n'
          'neg/foo.wav,MicA\n'
          'neg/bar.wav,MicB\n'
          'neg/baz.wav,MicB\n'
      )

    aduio_sources = source_info.AudioSources(
        audio_globs=(
            source_info.AudioSourceConfig(
                dataset_name='test',
                base_path=self.tempdir,
                file_glob='*/*.wav',
                min_audio_len_s=0.0,
                target_sample_rate_hz=16000,
            ),
        )
    )

    placeholder_model_config = config_dict.ConfigDict()
    placeholder_model_config.embedding_size = 32
    placeholder_model_config.sample_rate = 16000
    model_config = embed.ModelConfig(
        model_key='placeholder_model',
        embedding_dim=32,
        model_config=placeholder_model_config,
    )

    with self.subTest('embedding'):
      embedding_dim = 32
      if db_key == 'in_mem':
        in_mem_db_config = config_dict.ConfigDict()
        in_mem_db_config.embedding_dim = embedding_dim
        db_config = db_loader.DBConfig(
            db_key='in_mem',
            db_config=in_mem_db_config,
        )
        db = db_config.load_db()
      elif db_key == 'sqlite_usearch':
        db_path = self.db_path + '_embedding'
        db = db_loader.create_new_usearch_db(
            db_path=db_path, embedding_dim=embedding_dim
        )
      else:
        raise ValueError(f'Unknown db_key: {db_key}')

      embed_worker = embed.EmbedWorker(
          audio_sources=aduio_sources,
          model_config=model_config,
          db=db,
      )
      embed_worker.process_all()
      # The hop size is 1.0s and each file is 6.0s, so we get 6 embeddings
      # per file. There are six files, so we should get 36 embeddings.
      self.assertEqual(db.count_embeddings(), 36)
      embs = db.get_embeddings_batch(db.match_window_ids())
      self.assertEqual(embs.shape[-1], 32)
      self.assertLen(db.get_all_deployments(), 2)
      self.assertLen(db.get_all_recordings(), 6)

      # Check that metadata got attached to deployments and recordings.
      deployments = db.get_all_deployments()
      for d in deployments:
        if d.name == 'pos':
          self.assertEqual(d.habitat, 'forest')
        elif d.name == 'neg':
          self.assertEqual(d.habitat, 'grassland')
      recordings = db.get_all_recordings()
      for r in recordings:
        if r.filename == 'pos/foo.wav':
          self.assertEqual(r.mic_type, 'MicA')
        elif r.filename == 'pos/baz.wav':
          self.assertEqual(r.mic_type, 'MicB')

      # Check that the metadata is set correctly.
      got_md = db.get_metadata(key=None)
      self.assertIn('audio_sources', got_md)
      self.assertIn('model_config', got_md)

    with self.subTest('labels'):
      embedding_dim = 6
      if db_key == 'in_mem':
        in_mem_db_config = config_dict.ConfigDict()
        in_mem_db_config.embedding_dim = embedding_dim
        db_config = db_loader.DBConfig(
            db_key='in_mem',
            db_config=in_mem_db_config,
        )
        db = db_config.load_db()
      elif db_key == 'sqlite_usearch':
        db_path = self.db_path + '_labels'
        db = db_loader.create_new_usearch_db(
            db_path=db_path, embedding_dim=embedding_dim
        )
      else:
        raise ValueError(f'Unknown db_key: {db_key}')

      model_config.logits_key = 'label'
      model_config.logits_idxes = (1, 2, 3, 5, 8, 13)

      embed_worker = embed.EmbedWorker(
          audio_sources=aduio_sources,
          model_config=model_config,
          db=db,
      )
      embed_worker.process_all()
      # The hop size is 1.0s and each file is 6.0s, so we get 6 embeddings
      # per file. There are six files, so we should get 36 embeddings.
      self.assertEqual(db.count_embeddings(), 36)
      embs = db.get_embeddings_batch(db.match_window_ids())
      # The placeholder model defaults to 128-dim'l outputs, but we only want
      # the channels specified in the logits_idxes.
      self.assertEqual(embs.shape[-1], 6)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for linear classifier implementation."""

import os
import tempfile

from ml_collections import config_dict
import numpy as np
import pandas as pd
from perch_hoplite.agile import classifier
from perch_hoplite.agile import classifier_data
from perch_hoplite.db.tests import test_utils as db_test_utils

from absl.testing import absltest


class ClassifierTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    # `self.create_tempdir()` raises an UnparsedFlagAccessError, which is why
    # we use `tempdir` directly.
    self.tempdir = tempfile.mkdtemp()

  def _make_linear_classifier(self, embedding_dim, classes):
    np.random.seed(1234)
    beta = np.float32(np.random.normal(size=(embedding_dim, len(classes))))
    beta_bias = np.float32(np.random.normal(size=(len(classes),)))
    embedding_model_config = config_dict.ConfigDict({
        'model_name': 'nelson',
    })
    return classifier.LinearClassifier(
        beta, beta_bias, classes, embedding_model_config
    )

  def test_call_linear_classifier(self):
    embedding_dim = 8
    classes = ('a', 'b', 'c')
    classy = self._make_linear_classifier(embedding_dim, classes)

    batch_embeddings = np.random.normal(size=(10, embedding_dim))
    predictions = classy(batch_embeddings)
    self.assertEqual(predictions.shape, (10, len(classes)))

    single_embedding = np.random.normal(size=(embedding_dim,))
    predictions = classy(single_embedding)
    self.assertEqual(predictions.shape, (len(classes),))

  def test_save_load_linear_classifier(self):
    embedding_dim = 8
    classes = ('a', 'b', 'c')
    classy = self._make_linear_classifier(embedding_dim, classes)
    classy_path = os.path.join(self.tempdir, 'classifier.json')
    classy.save(classy_path)
    classy_loaded = classifier.LinearClassifier.load(classy_path)
    np.testing.assert_allclose(classy_loaded.beta, classy.beta)
    np.testing.assert_allclose(classy_loaded.beta_bias, classy.beta_bias)
    self.assertSequenceEqual(classy_loaded.classes, classy.classes)
    self.assertEqual(classy_loaded.embedding_model_config.model_name, 'nelson')

  def test_train_linear_classifier(self):
    rng = np.random.default_rng(1234)
    embedding_dim = 8
    db = db_test_utils.make_db(
        path=self.tempdir,
        db_type='in_mem',
        num_embeddings=1024,
        rng=rng,
        embedding_dim=embedding_dim,
    )
    db_test_utils.add_random_labels(
        db, rng=rng, unlabeled_prob=0.5, positive_label_prob=0.1
    )
    data_manager = classifier_data.AgileDataManager(
        target_labels=db_test_utils.CLASS_LABELS,
        db=db,
        train_ratio=0.8,
        min_eval_examples=5,
        batch_size=32,
        weak_negatives_batch_size=10,
        rng=np.random.default_rng(42),
    )
    lc, eval_scores = classifier.train_linear_classifier(
        data_manager,
        learning_rate=0.01,
        weak_neg_weight=0.5,
        num_train_steps=128,
        loss='bce',
    )
    self.assertIsInstance(lc, classifier.LinearClassifier)
    np.testing.assert_equal(
        lc.beta.shape, (embedding_dim, len(db_test_utils.CLASS_LABELS))
    )
    np.testing.assert_equal(
        lc.beta_bias.shape, (len(db_test_utils.CLASS_LABELS),)
    )
    self.assertIn('roc_auc', eval_scores)

  def test_write_inference_csv(self):
    embedding_dim = 8
    rng = np.random.default_rng(1234)
    db = db_test_utils.make_db(
        path=self.tempdir,
        db_type='in_mem',
        num_embeddings=1024,
        rng=rng,
        embedding_dim=embedding_dim,
    )
    db_test_utils.add_random_labels(
        db, rng=rng, unlabeled_prob=0.5, positive_label_prob=0.1
    )
    classes = ('alpha', 'beta', 'delta', 'epsilon')
    classy = self._make_linear_classifier(embedding_dim, classes)
    inference_classes = ('alpha', 'epsilon', 'gamma')
    classy.beta_bias = 0.0
    csv_filepath = os.path.join(self.tempdir, 'inference.csv')
    classifier.write_inference_csv(
        window_ids=np.array(db.match_window_ids()),
        linear_classifier=classy,
        db=db,
        output_filepath=csv_filepath,
        threshold=0.0,
        labels=inference_classes,
    )
    inference_csv = pd.read_csv(csv_filepath)
    got_labels = np.unique(inference_csv['label'].values)
    # `gamma` is not in the inference_classes, so should not be in the output.
    expected_labels = ('alpha', 'epsilon')
    np.testing.assert_array_equal(got_labels, expected_labels)

    # We can estimate the total number of detections. There are 1024 embeddings,
    # and we will only have outputs for two classes. Each logit is > 0 with
    # probability 0.5, because we are using an unbiased random classifier
    # with random embeddings. So, we expect 1024 * 0.5 * 2 = 1024 detections.
    self.assertGreater(len(inference_csv), 1000)
    self.assertLess(len(inference_csv), 1050)

    # Spot check some of the inference scores.
    for i in range(16):
      emb_id = inference_csv['idx'][i]
      lbl = inference_csv['label'][i]
      got_logit = inference_csv['logits'][i]
      class_idx = classy.classes.index(lbl)
      embedding = db.get_embedding(emb_id)
      expect_logit = classy(embedding)[class_idx]
      self.assertEqual(np.float16(got_logit), np.float16(expect_logit))


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Perch Taxonomy Model class."""

import dataclasses
from typing import Any

from absl import logging
from etils import epath
from ml_collections import config_dict
import numpy as np
from perch_hoplite.taxonomy import namespace
from perch_hoplite.zoo import hub
from perch_hoplite.zoo import zoo_interface
import tensorflow as tf


@dataclasses.dataclass
class TaxonomyModelTF(zoo_interface.EmbeddingModel):
  """Taxonomy SavedModel.

  Attributes:
    model_path: Path to model files.
    window_size_s: Window size for framing audio in seconds. TODO(tomdenton):
      Ideally this should come from a model metadata file.
    hop_size_s: Hop size for inference.
    model: Loaded TF SavedModel.
    class_list: Loaded class_list for the model's output logits.
    batchable: Whether the model supports batched input.
    target_peak: Peak normalization value.
  """

  model_path: str
  window_size_s: float
  hop_size_s: float
  model: Any  # TF SavedModel
  class_list: dict[str, namespace.ClassList]
  batchable: bool
  target_peak: float | None = 0.25
  tfhub_path: str | None = None
  tfhub_version: int | None = None

  @classmethod
  def is_batchable(cls, model: Any) -> bool:
    sig = model.signatures['serving_default']
    return sig.inputs[0].shape[0] is None

  @classmethod
  def load_class_lists(cls, csv_glob):
    class_lists = {}
    for csv_path in csv_glob:
      key = csv_path.name.replace('.csv', '')
      with csv_path.open('r') as f:
        try:
          class_lists[key] = namespace.ClassList.from_csv(f)
        except ValueError as e:
          # Some CSV assets aren't really class lists, so we just skip them.
          logging.warning('Failed to load class list %s: %s', csv_path, e)
          continue
    return class_lists

  @classmethod
  def from_tfhub(
      cls,
      config: config_dict.ConfigDict,
  ) -> 'TaxonomyModelTF':
    if hasattr(config, 'tfhub_path') and config.tfhub_path is not None:
      tfhub_path = config.tfhub_path
    else:
      raise ValueError('tfhub_path is required to load from TFHub.')
    if config.model_path:
      raise ValueError(
          'Exactly one of tfhub_version and model_path should be set.'
      )

    # This model behaves exactly like the usual saved_model.
    model = hub.load(tfhub_path, config.tfhub_version)

    # Check whether the model support polymorphic batch shape.
    batchable = cls.is_batchable(model)

    # Get the labels CSV from TFHub.
    model_path = hub.resolve(tfhub_path, config.tfhub_version)
    class_lists_glob = (epath.Path(model_path) / 'assets').glob('*.csv')
    class_lists = cls.load_class_lists(class_lists_glob)
    mutable_config = config.copy_and_resolve_references()
    del mutable_config.model_path
    return cls(
        model=model,
        class_list=class_lists,
        batchable=batchable,
        model_path=model_path,
        **mutable_config,
    )

  @classmethod
  def load_version(
      cls, tfhub_version: int, hop_size_s: float = 5.0
  ) -> 'TaxonomyModelTF':
    cfg = config_dict.ConfigDict({
        'model_path': '',
        'sample_rate': 32000,
        'window_size_s': 5.0,
        'hop_size_s': hop_size_s,
        'target_peak': 0.25,
        'tfhub_version': tfhub_version,
    })
    return cls.from_tfhub(cfg)

  @classmethod
  def load_surfperch_version(
      cls, tfhub_version: int, hop_size_s: float = 5.0
  ) -> 'TaxonomyModelTF':
    """Load a model from TFHub."""
    cfg = config_dict.ConfigDict({
        'model_path': '',
        'sample_rate': 32000,
        'window_size_s': 5.0,
        'hop_size_s': hop_size_s,
        'target_peak': 0.25,
        'tfhub_version': tfhub_version,
        'tfhub_path': hub.SURFPERCH_SLUG,
    })
    return cls.from_tfhub(cfg)

  @classmethod
  def load_v2_version(
      cls, tfhub_version: int, hop_size_s: float = 5.0
  ) -> 'TaxonomyModelTF':
    """Load a model from TFHub."""
    cfg = config_dict.ConfigDict({
        'model_path': '',
        'sample_rate': 32000,
        'window_size_s': 5.0,
        'hop_size_s': hop_size_s,
        'target_peak': 0.25,
        'tfhub_version': tfhub_version,
        'tfhub_path': hub.PERCH_V2_SLUG,
    })
    return cls.from_tfhub(cfg)

  @classmethod
  def from_config(cls, config: config_dict.ConfigDict) -> 'TaxonomyModelTF':
    logging.info('Loading taxonomy model...')

    if hasattr(config, 'tfhub_version') and config.tfhub_version is not None:
      return cls.from_tfhub(config)

    base_path = epath.Path(config.model_path)
    if (base_path / 'saved_model.pb').exists() and (
        base_path / 'assets'
    ).exists():
      # This looks like a downloaded TFHub model.
      model_path = base_path
      class_lists_glob = (epath.Path(model_path) / 'assets').glob('*.csv')
    else:
      # Probably a savedmodel distributed directly.
      model_path = base_path / 'savedmodel'
      class_lists_glob = epath.Path(base_path).glob('*.csv')

    model = tf.saved_model.load(model_path)
    class_lists = cls.load_class_lists(class_lists_glob)

    # Check whether the model support polymorphic batch shape.
    batchable = cls.is_batchable(model)
    return cls(
        model=model, class_list=class_lists, batchable=batchable, **config
    )

  def get_classifier_head(self, classes: list[str]):
    """Extract a classifier head for the desired subset of classes."""
    base_path = epath.Path(self.model_path)
    if (base_path / 'variables').exists():
      vars_filepath = f'{self.model_path}/variables/variables'
    elif (base_path / 'savedmodel' / 'variables').exists():
      vars_filepath = f'{self.model_path}/savedmodel/variables/variables'
    else:
      raise ValueError(f'No variables found in {self.model_path}')

    def _get_weights_and_bias(num_classes: int):
      weights = None
      bias = None
      for vname, vshape in tf.train.list_variables(vars_filepath):
        if len(vshape) == 1 and vshape[-1] == num_classes:
          if bias is None:
            bias = tf.train.load_variable(vars_filepath, vname)
          else:
            raise ValueError('Multiple possible biases for class list.')
        if len(vshape) == 2 and vshape[-1] == num_classes:
          if weights is None:
            weights = tf.train.load_variable(vars_filepath, vname)
          else:
            raise ValueError('Multiple possible weights for class list.')
      if hasattr(weights, 'numpy'):
        weights = weights.numpy()
      if hasattr(bias, 'numpy'):
        bias = bias.numpy()
      return weights, bias

    class_wts = {}
    for logit_key in self.class_list:
      num_classes = len(self.class_list[logit_key].classes)
      weights, bias = _get_weights_and_bias(num_classes)
      if weights is None or bias is None:
        raise ValueError(
            f'No weights or bias found for {logit_key} {num_classes}'
        )
      for i, k in enumerate(self.class_list[logit_key].classes):
        class_wts[k] = weights[:, i], bias[i]

    wts = []
    biases = []
    found_classes = []
    for target_class in classes:
      if target_class not in class_wts:
        continue
      wts.append(class_wts[target_class][0])
      biases.append(class_wts[target_class][1])
      found_classes.append(target_class)
    print(f'Found classes: {found_classes}')
    return found_classes, np.stack(wts, axis=-1), np.stack(biases, axis=-1)

  def embed(self, audio_array: np.ndarray) -> zoo_interface.InferenceOutputs:
    return zoo_interface.embed_from_batch_embed_fn(
        self.batch_embed, audio_array
    )

  def _nonbatchable_batch_embed(self, audio_batch: np.ndarray):
    """Embed a batch of audio with an old non-batchable model."""
    all_logits = []
    all_embeddings = []
    for audio in audio_batch:
      outputs = self.model.infer_tf(audio[np.newaxis, :])
      if hasattr(outputs, 'keys'):
        embedding = outputs.pop('embedding')
        logits = outputs.pop('label')
      else:
        # Assume the model output is always a (logits, embedding) twople.
        logits, embedding = outputs
      all_logits.append(logits)
      all_embeddings.append(embedding)
    all_logits = np.stack(all_logits, axis=0)
    all_embeddings = np.stack(all_embeddings, axis=0)
    return {
        'embedding': all_embeddings,
        'label': all_logits,
    }

  def batch_embed(
      self, audio_batch: np.ndarray[Any, Any]
  ) -> zoo_interface.InferenceOutputs:
    framed_audio = self.frame_audio(
        audio_batch, self.window_size_s, self.hop_size_s
    )
    framed_audio = self.normalize_audio(framed_audio, self.target_peak)
    rebatched_audio = framed_audio.reshape([-1, framed_audio.shape[-1]])

    if not self.batchable:
      outputs = self._nonbatchable_batch_embed(rebatched_audio)
    elif hasattr(self.model, 'infer_tf'):
      # Older TFJax export.
      outputs = self.model.infer_tf(rebatched_audio)
    else:
      # Perch v2 style.
      outputs = self.model.signatures['serving_default'](inputs=rebatched_audio)

    frontend_output = None
    if hasattr(outputs, 'keys'):
      # Dictionary-type outputs. Arrange appropriately.
      embeddings = outputs.pop('embedding')
      if 'frontend' in outputs:
        frontend_output = outputs.pop('frontend')
      elif 'spectrogram' in outputs:
        frontend_output = outputs.pop('spectrogram')
      if 'spatial_embedding' in outputs:
        outputs.pop('spatial_embedding')
      # Assume remaining outputs are all logits.
      logits = outputs
    elif len(outputs) == 2:
      # Assume logits, embeddings outputs.
      label_logits, embeddings = outputs
      logits = {'label': label_logits}
    else:
      raise ValueError('Unexpected outputs type.')

    for k, v in logits.items():
      logits[k] = np.reshape(v, framed_audio.shape[:2] + (v.shape[-1],))
    # Unbatch and add channel dimension.
    embeddings = np.reshape(
        embeddings,
        framed_audio.shape[:2]
        + (
            1,
            embeddings.shape[-1],
        ),
    )
    # Reshape frontend output to match embeddings.
    if frontend_output is not None:
      frontend_output = np.reshape(
          frontend_output,
          framed_audio.shape[:2] + frontend_output.shape[1:],
      )
    return zoo_interface.InferenceOutputs(
        embeddings=embeddings,
        logits=logits,
        separated_audio=None,
        batched=True,
        frontend=frontend_output,
    )



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Database configuration and constructor."""

import dataclasses

from etils import epath
from ml_collections import config_dict
from perch_hoplite.db import datatypes
from perch_hoplite.db import in_mem_impl
from perch_hoplite.db import interface
from perch_hoplite.db import sqlite_usearch_impl
import tqdm


@dataclasses.dataclass
class DBConfig(datatypes.HopliteConfig):
  """Configuration for embedding database.

  Attributes:
    db_key: Key for the database implementation to use.
    db_config: Configuration for the database implementation.
  """

  db_key: str
  db_config: config_dict.ConfigDict

  def load_db(self) -> interface.HopliteDBInterface:
    """Load the database from the specified path."""
    if self.db_key == 'sqlite_usearch':
      return sqlite_usearch_impl.SQLiteUSearchDB.create(**self.db_config)
    elif self.db_key == 'in_mem':
      return in_mem_impl.InMemoryGraphSearchDB.create(**self.db_config)
    else:
      raise ValueError(f'Unknown db_key: {self.db_key}')


def duplicate_db(
    source_db: interface.HopliteDBInterface,
    target_db_key: str,
    target_db_config: config_dict.ConfigDict,
):
  """Create a new DB and copy all data in source_db into it."""
  target_db = DBConfig(target_db_key, target_db_config).load_db()
  target_db.commit()

  # Check that the target_db is empty. If not, we'll have to do something more
  # sophisticated.
  if target_db.count_embeddings():
    raise ValueError('Target DB is not empty.')

  # Clone deployments and keep a map between source and target ids.
  deployment_id_mapping = {None: None}
  for deployment in tqdm.tqdm(source_db.get_all_deployments()):
    target_id = target_db.insert_deployment(**deployment.to_kwargs(skip=['id']))
    deployment_id_mapping[deployment.id] = target_id
  target_db.commit()

  # Clone recordings and keep a map between source and target ids.
  recording_id_mapping = {}
  for recording in tqdm.tqdm(source_db.get_all_recordings()):
    target_id = target_db.insert_recording(
        deployment_id=deployment_id_mapping[recording.deployment_id],
        **recording.to_kwargs(skip=['id', 'deployment_id']),
    )
    recording_id_mapping[recording.id] = target_id
  target_db.commit()

  # Clone windows and keep a map between source and target ids.
  window_id_mapping = {}
  for window in tqdm.tqdm(source_db.get_all_windows()):
    target_id = target_db.insert_window(
        recording_id=recording_id_mapping[window.recording_id],
        embedding=source_db.get_embedding(window.id),
        **window.to_kwargs(skip=['id', 'embedding', 'recording_id']),
    )
    window_id_mapping[window.id] = target_id
  target_db.commit()

  # Clone annotations and keep a map between source and target ids.
  annotation_id_mapping = {}
  for annotation in tqdm.tqdm(source_db.get_all_annotations()):
    target_id = target_db.insert_annotation(
        window_id=window_id_mapping[annotation.window_id],
        handle_duplicates='allow',
        **annotation.to_kwargs(skip=['id', 'window_id']),
    )
    annotation_id_mapping[annotation.id] = target_id
  target_db.commit()

  # Clone the KV store, replacing the DBConfig only.
  metadata = source_db.get_metadata(key=None)
  for k, v in metadata.items():
    if k == 'db_config':
      continue
    target_db.insert_metadata(k, v)
  target_db.insert_metadata('db_config', target_db_config)
  target_db.commit()

  return target_db, window_id_mapping


def create_new_usearch_db(
    db_path: str,
    embedding_dim: int,
) -> sqlite_usearch_impl.SQLiteUSearchDB:
  """Create a new USearch DB with the given path and embedding dimension."""
  epath.Path(db_path).parent.mkdir(parents=True, exist_ok=True)
  usearch_cfg = sqlite_usearch_impl.get_default_usearch_config(embedding_dim)
  return sqlite_usearch_impl.SQLiteUSearchDB.create(
      db_path=db_path, usearch_cfg=usearch_cfg
  )



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Iterator utilities."""

from collections.abc import Iterator

import numpy as np


def random_batched_iterator(
    ids: np.ndarray,
    batch_size: int,
    rng: np.random.RandomState,
) -> Iterator[np.ndarray]:
  """Yields batched embedding ids, shuffled after each of unlimited epochs."""
  if batch_size > len(ids):
    raise ValueError('Not enough ids to fill a batch.')
  rng.shuffle(ids)
  q = 0
  while True:
    if q + batch_size > len(ids):
      partial_batch = ids[q : len(ids)].copy()
      overflow = batch_size - len(partial_batch)
      rng.shuffle(ids)
      overflow_batch = ids[:overflow]
      batch = np.concatenate([partial_batch, overflow_batch], axis=0)
      q = overflow
    else:
      batch = ids[q : q + batch_size]
      q += batch_size
    yield batch.copy()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for namespace_db."""

import io
import tempfile

from absl import logging
from etils import epath
from perch_hoplite.taxonomy import namespace
from perch_hoplite.taxonomy import namespace_db

from absl.testing import absltest
from absl.testing import parameterized


class NamespaceDbTest(parameterized.TestCase):

  def test_load_namespace_db(self):
    db = namespace_db.load_db()

    # Check a couple ClassLists of known size.
    self.assertIn('caples', db.class_lists)
    caples_list = db.class_lists['caples']
    self.assertEqual(caples_list.namespace, 'ebird2021')
    self.assertLen(caples_list.classes, 79)

    genus_mapping = db.mappings['ebird2021_to_genus']
    caples_genera = caples_list.apply_namespace_mapping(genus_mapping)
    self.assertEqual(caples_genera.namespace, 'ebird2021_genera')
    self.assertLen(caples_genera.classes, 62)

    family_mapping = db.mappings['ebird2021_to_family']
    caples_families = caples_list.apply_namespace_mapping(family_mapping)
    self.assertEqual(caples_families.namespace, 'ebird2021_families')
    self.assertLen(caples_families.classes, 30)

    order_mapping = db.mappings['ebird2021_to_order']
    caples_orders = caples_list.apply_namespace_mapping(order_mapping)
    self.assertEqual(caples_orders.namespace, 'ebird2021_orders')
    self.assertLen(caples_orders.classes, 11)

  def test_class_map_csv(self):
    cl = namespace.ClassList(
        'ebird2021', ('amecro', 'amegfi', 'amered', 'amerob')
    )
    cl_csv = cl.to_csv()
    with io.StringIO(cl_csv) as f:
      got_cl = namespace.ClassList.from_csv(f)
    self.assertEqual(got_cl.namespace, 'ebird2021')
    self.assertEqual(got_cl.classes, ('amecro', 'amegfi', 'amered', 'amerob'))

    # Check that writing with tf.io.gfile behaves as expected, as newline
    # behavior may be different than working with StringIO.
    with tempfile.NamedTemporaryFile(suffix='.csv') as f:
      with epath.Path(f.name).open(mode='w') as gf:
        gf.write(cl_csv)
      with open(f.name, 'r') as f:
        got_cl = namespace.ClassList.from_csv(f.readlines())
    self.assertEqual(got_cl.namespace, 'ebird2021')
    self.assertEqual(got_cl.classes, ('amecro', 'amegfi', 'amered', 'amerob'))

  def test_namespace_class_list_closure(self):
    # Ensure that all classes in class lists appear in their namespace.
    db = namespace_db.load_db()

    all_missing_classes = set()
    for list_name, class_list in db.class_lists.items():
      missing_classes = set()
      namespace_ = db.namespaces[class_list.namespace]
      for cl in class_list.classes:
        if cl not in namespace_.classes:
          missing_classes.add(cl)
          all_missing_classes.add(cl)
      if missing_classes:
        logging.warning(
            'The classes %s in class list %s did not appear in namespace %s.',
            missing_classes,
            list_name,
            class_list.namespace,
        )
      missing_classes.discard('unknown')
    all_missing_classes.discard('unknown')
    self.assertEmpty(all_missing_classes)

  def test_namespace_mapping_closure(self):
    # Ensure that all classes in mappings appear in their namespace.
    db = namespace_db.load_db()

    all_missing_classes = set()
    for mapping_name, mapping in db.mappings.items():
      missing_source_classes = set()
      missing_target_classes = set()
      source_namespace = db.namespaces[mapping.source_namespace]
      target_namespace = db.namespaces[mapping.target_namespace]
      for source_cl, target_cl in mapping.mapped_pairs.items():
        if source_cl not in source_namespace.classes:
          missing_source_classes.add(source_cl)
          all_missing_classes.add(source_cl)
        if target_cl not in target_namespace.classes:
          missing_target_classes.add(target_cl)
          all_missing_classes.add(target_cl)
      if missing_source_classes:
        logging.warning(
            'The classes %s in mapping %s did not appear in namespace %s.',
            missing_source_classes,
            mapping_name,
            source_namespace.name,
        )
      if missing_target_classes:
        logging.warning(
            'The classes %s in mapping %s did not appear in namespace %s.',
            missing_target_classes,
            mapping_name,
            target_namespace.name,
        )
      missing_target_classes.discard('unknown')
    self.assertEmpty(all_missing_classes)

  def test_taxonomic_mappings(self):
    # Ensure that all ebird2021 species appear in taxonomic mappings.
    db = namespace_db.load_db()
    ebird = db.namespaces['ebird2021_species']
    genera = db.mappings['ebird2021_to_genus'].mapped_pairs
    families = db.mappings['ebird2021_to_family'].mapped_pairs
    orders = db.mappings['ebird2021_to_order'].mapped_pairs
    missing_genera = set()
    missing_families = set()
    missing_orders = set()
    for cl in ebird.classes:
      if cl not in genera:
        missing_genera.add(cl)
      if cl not in families:
        missing_families.add(cl)
      if cl not in orders:
        missing_orders.add(cl)
    self.assertEmpty(missing_genera)
    self.assertEmpty(missing_families)
    self.assertEmpty(missing_orders)


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for mass-embedding functionality."""

import os
import tempfile

from ml_collections import config_dict
import numpy as np
from perch_hoplite.taxonomy import namespace
from perch_hoplite.zoo import models_tf
from perch_hoplite.zoo import placeholder_model
from perch_hoplite.zoo import taxonomy_model_tf
import tensorflow as tf

from absl.testing import absltest
from absl.testing import parameterized


class ZooTest(parameterized.TestCase):

  def test_sep_embed_wrapper(self):
    """Check that the joint-model wrapper works as intended."""
    separator = placeholder_model.PlaceholderModel(
        sample_rate=22050,
        make_embeddings=False,
        make_logits=False,
        make_separated_audio=True,
    )

    embeddor = placeholder_model.PlaceholderModel(
        sample_rate=22050,
        make_embeddings=True,
        make_logits=True,
        make_separated_audio=False,
    )
    fake_config = config_dict.ConfigDict()
    sep_embed = models_tf.SeparateEmbedModel(
        sample_rate=22050,
        taxonomy_model_tf_config=fake_config,
        separator_model_tf_config=fake_config,
        separation_model=separator,
        embedding_model=embeddor,
    )
    audio = np.zeros(5 * 22050, np.float32)

    outputs = sep_embed.embed(audio)
    # The PlaceholderModel produces one embedding per second, and we have
    # five seconds of audio, with two separated channels, plus the channel
    # for the raw audio.
    # Note that this checks that the sample-rate conversion between the
    # separation model and embedding model has worked correctly.
    self.assertSequenceEqual(
        outputs.embeddings.shape, [5, 3, embeddor.embedding_size]
    )
    # The Sep+Embed model takes the max logits over the channel dimension.
    self.assertSequenceEqual(
        outputs.logits['label'].shape, [5, len(embeddor.class_list.classes)]
    )

  @parameterized.product(
      model_return_type=('tuple', 'dict'),
      batchable=(True, False),
  )
  def test_taxonomy_model_tf(self, model_return_type, batchable):
    class FakeModelFn:
      output_depths = {'label': 3, 'embedding': 256}
      spectrogram_size = (500, 128)

      def infer_tf(self, audio_array):
        outputs = {
            k: np.zeros([audio_array.shape[0], d], dtype=np.float32)
            for k, d in self.output_depths.items()
        }
        if batchable:
          outputs['spectrogram'] = np.zeros(
              [audio_array.shape[0], *self.spectrogram_size], dtype=np.float32
          )
        if model_return_type == 'tuple':
          # Published Perch models v1 through v4 returned a tuple, not a dict.
          return outputs['label'], outputs['embedding']
        return outputs

    class_list = {
        'label': namespace.ClassList('fake', ['alpha', 'beta', 'delta'])
    }
    wrapped_model = taxonomy_model_tf.TaxonomyModelTF(
        sample_rate=32000,
        model_path='/dev/null',
        window_size_s=5.0,
        hop_size_s=5.0,
        model=FakeModelFn(),
        class_list=class_list,
        batchable=batchable,
    )

    # Check that a single frame of audio is handled properly.
    outputs = wrapped_model.embed(np.zeros([5 * 32000], dtype=np.float32))
    self.assertFalse(outputs.batched)
    self.assertSequenceEqual(outputs.embeddings.shape, [1, 1, 256])
    self.assertSequenceEqual(outputs.logits['label'].shape, [1, 3])
    if batchable and model_return_type == 'dict':
      self.assertSequenceEqual(outputs.frontend.shape, [1, 500, 128])

    # Check that multi-frame audio is handled properly.
    outputs = wrapped_model.embed(np.zeros([20 * 32000], dtype=np.float32))
    self.assertFalse(outputs.batched)
    self.assertSequenceEqual(outputs.embeddings.shape, [4, 1, 256])
    self.assertSequenceEqual(outputs.logits['label'].shape, [4, 3])
    if batchable and model_return_type == 'dict':
      self.assertSequenceEqual(outputs.frontend.shape, [4, 500, 128])

    # Check that a batch of single frame of audio is handled properly.
    outputs = wrapped_model.batch_embed(
        np.zeros([10, 5 * 32000], dtype=np.float32)
    )
    self.assertTrue(outputs.batched)
    self.assertSequenceEqual(outputs.embeddings.shape, [10, 1, 1, 256])
    self.assertSequenceEqual(outputs.logits['label'].shape, [10, 1, 3])
    if batchable and model_return_type == 'dict':
      self.assertSequenceEqual(outputs.frontend.shape, [10, 1, 500, 128])

    # Check that a batch of multi-frame audio is handled properly.
    outputs = wrapped_model.batch_embed(
        np.zeros([2, 20 * 32000], dtype=np.float32)
    )
    self.assertTrue(outputs.batched)
    self.assertSequenceEqual(outputs.embeddings.shape, [2, 4, 1, 256])
    self.assertSequenceEqual(outputs.logits['label'].shape, [2, 4, 3])
    if batchable and model_return_type == 'dict':
      self.assertSequenceEqual(outputs.frontend.shape, [2, 4, 500, 128])

if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Conversion for TFRecord embeddings to Hoplite DB."""

from collections.abc import Sequence
import json

from etils import epath
from ml_collections import config_dict
import numpy as np
from perch_hoplite.agile import embed
from perch_hoplite.agile import source_info
from perch_hoplite.db import db_loader
from perch_hoplite.db import in_mem_impl
from perch_hoplite.db import sqlite_usearch_impl
import tensorflow as tf
import tqdm


def convert_tfrecords(
    embeddings_path: str,
    db_type: str,
    dataset_name: str,
    max_count: int = -1,
    **kwargs,
):
  """Convert a TFRecord embeddings dataset to a Hoplite DB."""
  ds = create_embeddings_dataset(
      embeddings_path,
      'embeddings-*',
  )
  # Peek at one embedding to get the embedding dimension.
  for ex in ds.as_numpy_iterator():
    emb_dim = ex['embedding'].shape[-1]
    break
  else:
    raise ValueError('No embeddings found.')

  if db_type == 'sqlite_usearch':
    db_path = kwargs['db_path']
    if (epath.Path(db_path) / sqlite_usearch_impl.HOPLITE_FILENAME).exists():
      raise ValueError(f'DB path {db_path} already exists.')
    db = db_loader.create_new_usearch_db(db_path, emb_dim)
  elif db_type == 'in_mem':
    db = in_mem_impl.InMemoryGraphSearchDB.create(embedding_dim=emb_dim)
  else:
    raise ValueError(f'Unknown db type: {db_type}')

  # Convert embedding config to new format and insert into the DB.
  legacy_config = load_embedding_config(embeddings_path)
  model_config = embed.ModelConfig(
      model_key=legacy_config.embed_fn_config.model_key,
      embedding_dim=emb_dim,
      model_config=legacy_config.embed_fn_config.model_config,
  )
  file_id_depth = legacy_config.embed_fn_config['file_id_depth']
  audio_globs = []
  for i, glob in enumerate(legacy_config.source_file_patterns):
    base_path, file_glob = glob.split('/')[-file_id_depth - 1 :]
    if i > 0:
      partial_dataset_name = f'{dataset_name}_{i}'
    else:
      partial_dataset_name = dataset_name
    audio_globs.append(
        source_info.AudioSourceConfig(
            dataset_name=partial_dataset_name,
            base_path=base_path,
            file_glob=file_glob,
            min_audio_len_s=legacy_config.embed_fn_config.min_audio_s,
            target_sample_rate_hz=legacy_config.embed_fn_config.get(
                'target_sample_rate_hz', -2
            ),
        )
    )

  audio_sources = source_info.AudioSources(audio_globs=tuple(audio_globs))
  db.insert_metadata('legacy_config', legacy_config)
  db.insert_metadata('audio_sources', audio_sources.to_config_dict())
  db.insert_metadata('model_config', model_config.to_config_dict())
  hop_size_s = model_config.model_config.hop_size_s
  window_size_s = model_config.model_config.window_size_s

  deployment_id = db.insert_deployment(name=dataset_name, project=dataset_name)
  recording_id_map = {}
  for ex in tqdm.tqdm(ds.as_numpy_iterator()):
    embs = ex['embedding']
    flat_embeddings = np.reshape(embs, [-1, embs.shape[-1]])
    file_id = str(ex['filename'], 'utf8')
    offset_s = ex['timestamp_s'].item()
    if max_count > 0 and db.count_embeddings() >= max_count:
      break
    # Long files were often split into shorter chunks during inference; ensure
    # consistent assignment of windows to recordings.
    if file_id in recording_id_map:
      recording_id = recording_id_map[file_id]
    else:
      recording_id = db.insert_recording(
          filename=file_id, deployment_id=deployment_id
      )
      recording_id_map[file_id] = recording_id
    windows_batch = [
        {
            'recording_id': recording_id,
            'offsets': [
                offset_s + hop_size_s * i,
                offset_s + hop_size_s * i + window_size_s,
            ],
        }
        for i in range(flat_embeddings.shape[0])
    ]
    db.insert_windows_batch(windows_batch, flat_embeddings)
    if max_count > 0 and db.count_embeddings() >= max_count:
      break
  db.commit()
  num_embeddings = db.count_embeddings()
  print('\n\nTotal embeddings : ', num_embeddings)
  hours_equiv = num_embeddings / 60 / 60 * hop_size_s
  print(f'\n\nHours of audio equivalent : {hours_equiv:.2f}')
  return db


# Functionality ported from chirp/inference/embed_lib.py and tf_examples.py.


def load_embedding_config(embeddings_path, filename: str = 'config.json'):
  """Loads the configuration to generate unlabeled embeddings."""
  embeddings_path = epath.Path(embeddings_path)
  with (embeddings_path / filename).open() as f:
    embedding_config = config_dict.ConfigDict(json.loads(f.read()))
  return embedding_config


# Feature keys.
FILE_NAME = 'filename'
TIMESTAMP_S = 'timestamp_s'
EMBEDDING = 'embedding'
EMBEDDING_SHAPE = 'embedding_shape'
LOGITS = 'logits'
SEPARATED_AUDIO = 'separated_audio'
SEPARATED_AUDIO_SHAPE = 'separated_audio_shape'
RAW_AUDIO = 'raw_audio'
RAW_AUDIO_SHAPE = 'raw_audio_shape'
FRONTEND = 'frontend'
FRONTEND_SHAPE = 'frontend_shape'


def get_feature_description(logit_names: Sequence[str] | None = None):
  """Create a feature description for the TFExamples.

  Each tensor feature includes both a serialized tensor and a 'shape' feature.
  The tensor feature can be parsed with tf.io.parse_tensor, and then reshaped
  according to the shape feature.

  Args:
    logit_names: Name of logit features included in the examples.

  Returns:
    Feature description dict for parsing TF Example protos.
  """
  feature_description = {
      FILE_NAME: tf.io.FixedLenFeature([], tf.string),
      TIMESTAMP_S: tf.io.FixedLenFeature([], tf.float32),
      EMBEDDING: tf.io.FixedLenFeature([], tf.string, default_value=''),
      EMBEDDING_SHAPE: tf.io.FixedLenSequenceFeature(
          [], tf.int64, allow_missing=True
      ),
      SEPARATED_AUDIO: tf.io.FixedLenFeature([], tf.string, default_value=''),
      SEPARATED_AUDIO_SHAPE: tf.io.FixedLenSequenceFeature(
          [], tf.int64, allow_missing=True
      ),
      FRONTEND: tf.io.FixedLenFeature([], tf.string, default_value=''),
      FRONTEND_SHAPE: tf.io.FixedLenSequenceFeature(
          [], tf.int64, allow_missing=True
      ),
      RAW_AUDIO: tf.io.FixedLenFeature([], tf.string, default_value=''),
      RAW_AUDIO_SHAPE: tf.io.FixedLenSequenceFeature(
          [], tf.int64, allow_missing=True
      ),
  }
  if logit_names is not None:
    for logit_name in logit_names:
      feature_description[logit_name] = tf.io.FixedLenFeature(
          [], tf.string, default_value=''
      )
      feature_description[f'{logit_name}_shape'] = (
          tf.io.FixedLenSequenceFeature([], tf.int64, allow_missing=True)
      )
  return feature_description


def get_example_parser(
    logit_names: Sequence[str] | None = None, tensor_dtype: str = 'float32'
):
  """Create a parser for decoding inference library TFExamples."""
  features = get_feature_description(logit_names=logit_names)

  def _parser(ex):
    ex = tf.io.parse_single_example(ex, features)
    tensor_keys = [EMBEDDING, SEPARATED_AUDIO, RAW_AUDIO, FRONTEND]
    if logit_names is not None:
      tensor_keys.extend(logit_names)
    for key in tensor_keys:
      # Note that we can't use implicit truthiness for string tensors.
      # We are also required to have the same tensor structure and dtype in
      # both conditional branches. So we use an empty tensor when no
      # data is present to parse.
      if ex[key] != tf.constant(b'', dtype=tf.string):
        ex[key] = tf.io.parse_tensor(ex[key], out_type=tensor_dtype)
      else:
        ex[key] = tf.zeros_like([], dtype=tensor_dtype)
    return ex

  return _parser


def create_embeddings_dataset(
    embeddings_dir,
    file_glob: str = '*',
    prefetch: int = 128,
    logit_names: Sequence[str] | None = None,
    tensor_dtype: str = 'float32',
    shuffle_files: bool = False,
):
  """Create a TF Dataset of the embeddings."""
  embeddings_dir = epath.Path(embeddings_dir)
  embeddings_files = [fn.as_posix() for fn in embeddings_dir.glob(file_glob)]
  if shuffle_files:
    np.random.shuffle(embeddings_files)
  ds = tf.data.TFRecordDataset(
      embeddings_files, num_parallel_reads=tf.data.AUTOTUNE
  )

  parser = get_example_parser(
      logit_names=logit_names, tensor_dtype=tensor_dtype
  )
  ds = ds.map(parser, num_parallel_calls=tf.data.AUTOTUNE)
  ds = ds.prefetch(prefetch)
  return ds



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SQLite database implementation using USearch for vector storage & search."""

import collections
from collections.abc import Sequence
import dataclasses
import datetime as dt
import functools
import itertools
import json
import re
import sqlite3
from typing import Any, Literal

from absl import logging
from etils import epath
from ml_collections import config_dict
import numpy as np
from perch_hoplite.db import brutalism
from perch_hoplite.db import datatypes
from perch_hoplite.db import interface
from perch_hoplite.db import score_functions
from perch_hoplite.db import search_results
from usearch import index as uindex

HOPLITE_FILENAME = 'hoplite.sqlite'
UINDEX_FILENAME = 'usearch.index'
USEARCH_CONFIG_KEY = 'usearch_config'
USEARCH_DTYPES = {
    'float16': uindex.ScalarKind.F16,
}
SQL_TYPE_TO_PYTHON_TYPE = {
    'INTEGER': int,
    'REAL': float,
    'TEXT': str,
    'BLOB': bytes,
    'FLOAT_LIST': list,
}
PYTHON_TYPE_TO_SQL_TYPE = {
    int: 'INTEGER',
    float: 'REAL',
    str: 'TEXT',
    bytes: 'BLOB',
    dt.datetime: 'TEXT',
    list: 'FLOAT_LIST',
}


def adapt_float_list(data: list[float]) -> bytes:
  return np.array(
      data,
      dtype=np.dtype('<f8'),  # little-endian np.float64
  ).tobytes()


def convert_float_list(blob: bytes) -> list[float]:
  return np.frombuffer(
      blob,
      dtype=np.dtype('<f8'),  # little-endian np.float64
  ).tolist()


def approx_float_list(blob: bytes, target: bytes) -> bool:
  return np.allclose(
      convert_float_list(blob),
      convert_float_list(target),
      rtol=0.0,
      atol=1e-6,
  )


sqlite3.register_adapter(list, adapt_float_list)
sqlite3.register_converter('FLOAT_LIST', convert_float_list)


def get_default_usearch_config(
    embedding_dim: int,
) -> config_dict.ConfigDict:
  """Get a default USearch config for the given embedding dimension."""
  usearch_cfg = config_dict.ConfigDict()
  usearch_cfg.embedding_dim = embedding_dim
  usearch_cfg.dtype = 'float16'
  usearch_cfg.metric_name = 'IP'
  usearch_cfg.expansion_add = 256
  usearch_cfg.expansion_search = 128
  return usearch_cfg


def is_valid_sql_identifier(name: str) -> bool:
  """Check if a string is a valid and safe SQL identifier."""

  if not name or not isinstance(name, str):
    return False

  # Regex to verify that the name starts with a letter or underscore, then
  # follows with letters, numbers or underscores.
  return re.match(r'^[a-zA-Z_][a-zA-Z0-9_]*$', name) is not None


def normalize_sql_value(value: Any) -> Any:
  """Normalize a python value to one of the types supported by SQL."""

  if (
      isinstance(value, list)
      or isinstance(value, tuple)
      or isinstance(value, np.ndarray)
  ):
    return [normalize_sql_value(v) for v in value]
  if isinstance(value, datatypes.LabelType):
    return value.value
  elif isinstance(value, dt.datetime):
    return value.isoformat()
  elif isinstance(value, np.integer):
    return int(value)
  elif isinstance(value, np.floating):
    return float(value)
  return value


def format_sql_insert_values(
    **kwargs: Any,
) -> tuple[str, str, list[Any]]:
  """Build columns string, placeholders string and values list for SQL INSERT.

  Args:
    **kwargs: Key-value pairs to pass to the SQL statement.

  Returns:
    A tuple of: a formatted columns string, a formatted placeholders string, and
    a list of corresponding values. Safe to be used in SQL INSERT statements.
  """

  for key in kwargs:
    if not is_valid_sql_identifier(key):
      raise ValueError(f'`{key}` is not a valid SQL identifier.')

  columns = list(kwargs.keys())
  placeholders = ['?'] * len(columns)
  values = normalize_sql_value(list(kwargs.values()))

  return f"({', '.join(columns)})", f"({', '.join(placeholders)})", values


def format_sql_update_on_conflict(*args: str) -> str:
  """Build the update part of ON CONFLICT clauses for INSERT statements."""

  for key in args:
    if not is_valid_sql_identifier(key):
      raise ValueError(f'`{key}` is not a valid SQL identifier.')

  if not args:
    return 'DO NOTHING'
  else:
    update_clauses_str = ', '.join([f'{key} = excluded.{key}' for key in args])
    return f'DO UPDATE SET {update_clauses_str}'


def format_sql_where_conditions(
    filter_dict: config_dict.ConfigDict | None = None,
    table_prefix: str | None = None,
) -> tuple[str, list[Any]]:
  r"""Build conditions string and values list for SQL WHERE from given filters.

  Args:
    filter_dict: A ConfigDict of constraints to build SQL conditions from.
    table_prefix: An optional table prefix to prepend to each column name.

  Returns:
    A tuple of: a formatted string of AND-joined conditions, and a list of
    corresponding values. Safe to be used in SQL WHERE statements.
  """

  if table_prefix and not is_valid_sql_identifier(table_prefix):
    raise ValueError(
        f'Table prefix `{table_prefix}` is not a valid SQL identifier.'
    )

  if not filter_dict:
    return '', []

  supported_ops = {
      'eq',
      'neq',
      'lt',
      'lte',
      'gt',
      'gte',
      'isin',
      'notin',
      'range',
      'approx',
  }

  conditions = []
  values = []

  # Build the SQL conditions for each operation.
  for op_name, op_filters in filter_dict.items():

    if op_name not in supported_ops:
      raise ValueError(
          f'Unsupported operation: `{op_name}`. Supported filtering operations'
          f' are: {supported_ops}.'
      )
    if not isinstance(op_filters, config_dict.ConfigDict):
      raise ValueError(f'`{op_name}` value must be a ConfigDict.')

    for key, value in op_filters.items():
      if table_prefix:
        column = f'{table_prefix}.{key}'
      else:
        column = key

      # Check that the key is a valid SQL identifier.
      if not is_valid_sql_identifier(key):
        raise ValueError(
            f'Table column `{column}` is not a valid SQL identifier. Fix the'
            f' filter rule under the `{op_name}` operation.'
        )

      # Normalize the value.
      value = normalize_sql_value(value)

      # Build the current SQL condition.
      if op_name == 'eq':
        if key == 'offsets':
          logging.warning(
              "Do not apply `eq` to the `offsets` unless you know what you're "
              'doing. Apply `approx` instead to avoid floating point errors.'
          )
        if value is None:
          conditions.append(f'{column} IS NULL')
        else:
          conditions.append(f'{column} = ?')
          values.append(value)
      elif op_name == 'neq':
        if value is None:
          conditions.append(f'{column} IS NOT NULL')
        else:
          conditions.append(f'{column} != ?')
          values.append(value)
      elif op_name == 'lt':
        conditions.append(f'{column} < ?')
        values.append(value)
      elif op_name == 'lte':
        conditions.append(f'{column} <= ?')
        values.append(value)
      elif op_name == 'gt':
        conditions.append(f'{column} > ?')
        values.append(value)
      elif op_name == 'gte':
        conditions.append(f'{column} >= ?')
        values.append(value)
      elif op_name == 'isin':
        if not isinstance(value, list):
          raise ValueError(f'`{op_name}` value must be a list.')
        placeholders = ['?'] * len(value)
        placeholders_str = ', '.join(placeholders)
        conditions.append(f'{column} IN ({placeholders_str})')
        values.extend(value)
      elif op_name == 'notin':
        if not isinstance(value, list):
          raise ValueError(f'`{op_name}` value must be a list.')
        placeholders = ['?'] * len(value)
        placeholders_str = ', '.join(placeholders)
        conditions.append(f'{column} NOT IN ({placeholders_str})')
        values.extend(value)
      elif op_name == 'range':
        if not isinstance(value, list) or len(value) != 2:
          raise ValueError(f'`{op_name}` value must be a list of 2 elements.')
        conditions.append(f'{column} BETWEEN ? AND ?')
        values.extend(value)
      elif op_name == 'approx':
        if key == 'offsets':
          conditions.append(f'APPROX_FLOAT_LIST({column}, ?) = TRUE')
        else:
          conditions.append(f'ABS({column} - ?) < 1e-6')
        values.append(value)

  return ' AND '.join(conditions), values


@dataclasses.dataclass
class SQLiteUSearchDB(interface.HopliteDBInterface):
  """SQLite hoplite database, using USearch for vector storage.

  USearch provides both indexing for approximate nearest neighbor search and
  fast disk-based random access to vectors for the complete database. USearch
  will default to working with disk-based vectors, unless we insert or remove
  embeddings, in which case we load the index into memory and use it from there
  for all subsequent operations. On database commit, the in-memory index is
  persisted to disk.

  Attributes:
    db_path: The path to the database directory.
    db: The sqlite3 database connection.
    ui: The USearch index.
    _embedding_dim: The dimension of the embeddings.
    _embedding_dtype: The data type of the embeddings.
    _cursor: The sqlite3 cursor.
    _ui_loaded: Whether the USearch index was loaded in memory.
    _ui_updated: Whether the USearch index was updated since the last load and
      needs to be persisted to disk at some point in the future.
  """

  # User-provided.
  db_path: epath.Path

  # Instantiated during creation.
  db: sqlite3.Connection
  ui: uindex.Index

  # Obtained from `usearch_cfg`.
  _embedding_dim: int
  _embedding_dtype: type[Any] = np.float16

  # Dynamic state.
  _cursor: sqlite3.Cursor | None = None
  _ui_loaded: bool = False
  _ui_updated: bool = False

  @property
  def sqlite_path(self) -> epath.Path:
    return self.db_path / HOPLITE_FILENAME

  @property
  def usearch_path(self) -> epath.Path:
    return self.db_path / UINDEX_FILENAME

  @staticmethod
  def _setup_tables(cursor: sqlite3.Cursor) -> None:
    """Create the SQLite tables.

    Args:
      cursor: The SQLite cursor to use.
    """

    # Skip setting up the tables if they already exist.
    cursor.execute("""
        SELECT name
        FROM sqlite_master
        WHERE name = "windows" AND type = "table"
        """)
    if cursor.fetchone() is not None:
      return

    # Enable foreign keys.
    cursor.execute('PRAGMA foreign_keys = ON')

    # Create the metadata table.
    cursor.execute("""
        CREATE TABLE IF NOT EXISTS hoplite_metadata (
            key TEXT PRIMARY KEY,
            value TEXT NOT NULL
        )
        """)

    # Create the deployments table.
    cursor.execute("""
        CREATE TABLE IF NOT EXISTS deployments (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            name TEXT NOT NULL,
            project TEXT NOT NULL,
            latitude REAL,
            longitude REAL,
            UNIQUE (name, project)
        )
        """)

    # Create the recordings table.
    cursor.execute("""
        CREATE TABLE IF NOT EXISTS recordings (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            filename TEXT NOT NULL,
            datetime TEXT,
            deployment_id INTEGER REFERENCES deployments(id) ON DELETE CASCADE,
            UNIQUE (id, deployment_id),
            UNIQUE (filename, deployment_id)
        )
        """)

    # Create the windows table.
    cursor.execute("""
        CREATE TABLE IF NOT EXISTS windows (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            recording_id INTEGER NOT NULL REFERENCES recordings(id) ON DELETE CASCADE,
            offsets FLOAT_LIST NOT NULL,
            UNIQUE (id, recording_id, offsets)
        )
        """)

    # Create the annotations table.
    cursor.execute("""
        CREATE TABLE IF NOT EXISTS annotations (
            id INTEGER PRIMARY KEY AUTOINCREMENT,
            recording_id INTEGER NOT NULL REFERENCES recordings(id) ON DELETE CASCADE,
            offsets FLOAT_LIST NOT NULL,
            label TEXT NOT NULL,
            label_type INTEGER NOT NULL,
            provenance TEXT NOT NULL,
            UNIQUE (id, recording_id, offsets)
        )
        """)

    # Create other indexes for efficient lookups.
    cursor.execute("""
        CREATE INDEX IF NOT EXISTS idx_annotations
        ON annotations(recording_id, offsets, label, label_type, provenance)
        """)
    cursor.execute("""
        CREATE INDEX IF NOT EXISTS idx_labels
        ON annotations(label, label_type, provenance)
        """)

  @staticmethod
  def _get_all_metadata(cursor: sqlite3.Cursor) -> config_dict.ConfigDict:
    """Get all key-value pairs from the metadata table.

    Args:
      cursor: The SQLite cursor to use.

    Returns:
      A ConfigDict containing all the metadata.
    """

    cursor.execute("""
        SELECT key, value
        FROM hoplite_metadata
        """)
    return config_dict.ConfigDict(
        {k: json.loads(v) for k, v in cursor.fetchall()}
    )

  @classmethod
  def create(
      cls,
      db_path: str,
      usearch_cfg: config_dict.ConfigDict | None = None,
  ) -> 'SQLiteUSearchDB':
    """Connect to and, if needed, initialize the database.

    Args:
      db_path: The path to the database directory.
      usearch_cfg: The configuration for the USearch index. If None, the config
        is loaded from the DB.

    Returns:
      A new instance of the database.
    """

    # Create the SQLite DB.
    db_path = epath.Path(db_path)
    db_path.mkdir(parents=True, exist_ok=True)
    sqlite_path = db_path / HOPLITE_FILENAME
    db = sqlite3.connect(
        sqlite_path.as_posix(),
        detect_types=sqlite3.PARSE_DECLTYPES,
    )
    db.create_function(
        name='APPROX_FLOAT_LIST',
        narg=2,
        func=approx_float_list,
        deterministic=True,
    )
    db.set_trace_callback(
        lambda statement: logging.info('Executed SQL statement: %s', statement)
    )
    cursor = db.cursor()
    cursor.execute('PRAGMA journal_mode = WAL')  # Enable WAL mode.
    cls._setup_tables(cursor)
    db.commit()

    # Retrieve the metadata.
    # TODO(tomdenton): Check that `usearch_cfg` is consistent with the DB.
    metadata = cls._get_all_metadata(cursor)
    if (
        USEARCH_CONFIG_KEY in metadata
        and usearch_cfg is not None
        and metadata[USEARCH_CONFIG_KEY] != usearch_cfg
    ):
      raise ValueError(
          'A usearch_cfg was provided, but a different one already exists in'
          ' the DB.'
      )
    if USEARCH_CONFIG_KEY in metadata:
      usearch_cfg = metadata[USEARCH_CONFIG_KEY]
    elif usearch_cfg is None:
      raise ValueError('No usearch_cfg was found in DB and none was provided.')

    # Create the USearch index.
    usearch_dtype = USEARCH_DTYPES[usearch_cfg.dtype]
    index_path = db_path / UINDEX_FILENAME
    if index_path.exists():
      ui = uindex.Index(
          ndim=usearch_cfg.embedding_dim,
          path=index_path,
          view=True,
      )
      ui_in_memory = False
    else:
      ui = uindex.Index(
          ndim=usearch_cfg.embedding_dim,
          metric=getattr(uindex.MetricKind, usearch_cfg.metric_name),
          expansion_add=usearch_cfg.expansion_add,
          expansion_search=usearch_cfg.expansion_search,
          dtype=usearch_dtype,
          path=index_path,
          view=False,
      )
      ui_in_memory = True

    # Create the Hoplite DB.
    hoplite_db = cls(
        db_path=db_path,
        db=db,
        ui=ui,
        _embedding_dim=usearch_cfg.embedding_dim,
        _embedding_dtype=usearch_cfg.dtype,
        _ui_loaded=ui_in_memory,
        _ui_updated=ui_in_memory,
    )
    hoplite_db.insert_metadata(USEARCH_CONFIG_KEY, usearch_cfg)
    hoplite_db.commit()
    return hoplite_db

  def add_extra_table_column(
      self,
      table_name: str,
      column_name: str,
      column_type: type[Any],
  ) -> None:
    """Add an extra column to a table in the database."""

    if table_name not in [
        'deployments',
        'recordings',
        'windows',
        'annotations',
    ]:
      raise ValueError(f'Table `{table_name}` does not exist.')
    if not is_valid_sql_identifier(column_name):
      raise ValueError(f'Column `{column_name}` is not a valid SQL identifier.')
    if not isinstance(column_type, type):
      raise ValueError(f'Column type `{column_type}` must be a type.')

    if column_type not in PYTHON_TYPE_TO_SQL_TYPE:
      raise ValueError(
          f'Column type `{column_type.__name__}` is not supported. Use one of:'
          f' {", ".join([key.__name__ for key in PYTHON_TYPE_TO_SQL_TYPE.keys()])}'
      )

    cursor = self._get_cursor()
    cursor.execute(f"""
        ALTER TABLE {table_name}
        ADD COLUMN {column_name} {PYTHON_TYPE_TO_SQL_TYPE[column_type]}
        """)

    # Clear the cached property so that it is recomputed on the next access.
    self.__dict__.pop('_extra_table_columns', None)

  @functools.cached_property
  def _extra_table_columns(self) -> dict[str, dict[str, type[Any]]]:
    """Get all extra columns in the database."""
    tables = ['deployments', 'recordings', 'windows', 'annotations']
    default_columns = {
        'deployments': {'id', 'name', 'project', 'latitude', 'longitude'},
        'recordings': {'id', 'filename', 'datetime', 'deployment_id'},
        'windows': {'id', 'recording_id', 'offsets'},
        'annotations': {
            'id',
            'recording_id',
            'offsets',
            'label',
            'label_type',
            'provenance',
        },
    }
    extra_columns = {t: {} for t in tables}
    cursor = self._get_cursor()
    for table in tables:
      cursor.execute(f'PRAGMA table_info({table})')
      columns_info = cursor.fetchall()
      for col_info in columns_info:
        # col_info: cid, name, type, notnull, dflt_value, pk
        col_name = col_info[1]
        col_type = col_info[2]
        if col_name not in default_columns[table]:
          try:
            extra_columns[table][col_name] = SQL_TYPE_TO_PYTHON_TYPE[col_type]
          except KeyError as e:
            raise ValueError(
                f'Unsupported column type {col_type} for column '
                f'{col_name} in table {table}'
            ) from e
    return extra_columns

  def get_extra_table_columns(self) -> dict[str, dict[str, type[Any]]]:
    """Get all extra columns in the database."""
    return self._extra_table_columns

  def commit(self) -> None:
    """Commit any pending transactions to the database."""
    self.db.commit()
    if self._cursor is not None:
      self._cursor.close()
      self._cursor = None
    if self._ui_updated:
      self.ui.save()
      self._ui_updated = False

  def rollback(self) -> None:
    """Rollback any pending transactions to the database."""
    self.db.rollback()
    if self._cursor is not None:
      self._cursor.close()
      self._cursor = None

  def thread_split(self) -> 'SQLiteUSearchDB':
    """Get a new instance of the SQLite DB."""
    return self.create(self.db_path.as_posix())

  def _get_cursor(self) -> sqlite3.Cursor:
    """Get the SQLite cursor."""
    if self._cursor is None:
      self._cursor = self.db.cursor()
    return self._cursor

  def insert_metadata(self, key: str, value: config_dict.ConfigDict) -> None:
    """Insert a key-value pair into the metadata table."""

    cursor = self._get_cursor()
    json_coded = value.to_json()
    cursor.execute(
        """
        INSERT INTO hoplite_metadata (key, value)
        VALUES (?, ?)
        ON CONFLICT (key)
        DO UPDATE SET value = excluded.value
        """,
        (key, json_coded),
    )

  def get_metadata(self, key: str | None) -> config_dict.ConfigDict:
    """Get a key-value pair from the metadata table."""

    cursor = self._get_cursor()

    if key is None:
      cursor.execute("""
          SELECT key, value
          FROM hoplite_metadata
          """)
      return config_dict.ConfigDict(
          {k: json.loads(v) for k, v in cursor.fetchall()}
      )

    cursor.execute(
        """
        SELECT value
        FROM hoplite_metadata
        WHERE key = ?
        """,
        (key,),
    )
    result = cursor.fetchone()
    if result is None:
      raise KeyError(f'Metadata key not found: {key}')
    return config_dict.ConfigDict(json.loads(result[0]))

  def remove_metadata(self, key: str | None) -> None:
    """Remove a key-value pair from the metadata table."""

    cursor = self._get_cursor()

    if key is None:
      cursor.execute("""
          DELETE FROM hoplite_metadata
          """)
      return

    cursor.execute(
        """
        DELETE FROM hoplite_metadata
        WHERE key = ?
        """,
        (key,),
    )
    if cursor.rowcount == 0:
      raise KeyError(f'Metadata key not found: {key}')

  def insert_deployment(
      self,
      name: str,
      project: str,
      latitude: float | None = None,
      longitude: float | None = None,
      **kwargs: Any,
  ) -> int:
    """Insert a deployment into the database."""

    for key, value in kwargs.items():
      if key not in self._extra_table_columns['deployments']:
        self.add_extra_table_column('deployments', key, type(value))
    cursor = self._get_cursor()
    columns_str, placeholders_str, values = format_sql_insert_values(
        name=name,
        project=project,
        latitude=latitude,
        longitude=longitude,
        **kwargs,
    )
    update_clause_str = format_sql_update_on_conflict(
        'latitude',
        'longitude',
        *self._extra_table_columns['deployments'].keys(),
    )
    cursor.execute(
        f"""
        INSERT INTO deployments {columns_str}
        VALUES {placeholders_str}
        ON CONFLICT (name, project)
        {update_clause_str}
        """,
        values,
    )

    deployment_id = cursor.lastrowid
    if deployment_id is None:
      raise RuntimeError('Error inserting the deployment into the database.')
    return deployment_id

  def get_deployment(self, deployment_id: int) -> datatypes.Deployment:
    """Get a deployment from the database."""

    deployment_id = int(deployment_id)

    cursor = self._get_cursor()
    cursor.execute(
        """
        SELECT *
        FROM deployments
        WHERE id = ?
        """,
        (deployment_id,),
    )
    result = cursor.fetchone()
    if result is None:
      raise KeyError(f'Deployment id not found: {deployment_id}')

    columns = [col[0] for col in cursor.description]
    return datatypes.Deployment(**dict(zip(columns, result)))

  def remove_deployment(self, deployment_id: int) -> None:
    """Remove a deployment from the database."""

    deployment_id = int(deployment_id)

    remove_window_ids = self.match_window_ids(
        deployments_filter=config_dict.create(
            eq=dict(deployment_id=deployment_id)
        )
    )
    if remove_window_ids:
      if not self._ui_loaded:
        self.ui.load()
        self._ui_loaded = True
      self.ui.remove(remove_window_ids)
      self._ui_updated = True

    cursor = self._get_cursor()
    cursor.execute(
        """
        DELETE FROM deployments
        WHERE id = ?
        """,
        (deployment_id,),
    )
    if cursor.rowcount == 0:
      raise KeyError(f'Deployment id not found: {deployment_id}')

  def insert_recording(
      self,
      filename: str,
      datetime: dt.datetime | None = None,
      deployment_id: int | None = None,
      **kwargs: Any,
  ) -> int:
    """Insert a recording into the database."""
    for key, value in kwargs.items():
      if key not in self._extra_table_columns['recordings']:
        self.add_extra_table_column('recordings', key, type(value))

    cursor = self._get_cursor()
    columns_str, placeholders_str, values = format_sql_insert_values(
        filename=filename,
        datetime=datetime,
        deployment_id=deployment_id,
        **kwargs,
    )
    update_clause_str = format_sql_update_on_conflict(
        'datetime',
        *self._extra_table_columns['recordings'].keys(),
    )
    try:
      cursor.execute(
          f"""
          INSERT INTO recordings {columns_str}
          VALUES {placeholders_str}
          ON CONFLICT (filename, deployment_id)
          {update_clause_str}
          """,
          values,
      )
    except sqlite3.Error as e:
      if e.sqlite_errorname == 'SQLITE_CONSTRAINT_FOREIGNKEY':
        custom_msg = 'Check that the deployment_id exists.'
      else:
        custom_msg = ''
      raise RuntimeError(
          f'Error inserting the recording into the database. {custom_msg}'
      ) from e

    recording_id = cursor.lastrowid
    if recording_id is None:
      raise RuntimeError('Error inserting the recording into the database.')
    return recording_id

  def get_recording(self, recording_id: int) -> datatypes.Recording:
    """Get a recording from the database."""

    recording_id = int(recording_id)

    cursor = self._get_cursor()
    cursor.execute(
        """
        SELECT *
        FROM recordings
        WHERE id = ?
        """,
        (recording_id,),
    )
    result = cursor.fetchone()
    if result is None:
      raise KeyError(f'Recording id not found: {recording_id}')

    columns = [col[0] for col in cursor.description]
    recording = datatypes.Recording(**dict(zip(columns, result)))
    if recording.datetime is not None:
      recording.datetime = dt.datetime.fromisoformat(recording.datetime)
    return recording

  def remove_recording(self, recording_id: int) -> None:
    """Remove a recording from the database."""

    recording_id = int(recording_id)

    remove_window_ids = self.match_window_ids(
        recordings_filter=config_dict.create(eq=dict(recording_id=recording_id))
    )
    if remove_window_ids:
      if not self._ui_loaded:
        self.ui.load()
        self._ui_loaded = True
      self.ui.remove(remove_window_ids)
      self._ui_updated = True

    cursor = self._get_cursor()
    cursor.execute(
        """
        DELETE FROM recordings
        WHERE id = ?
        """,
        (recording_id,),
    )
    if cursor.rowcount == 0:
      raise KeyError(f'Recording id not found: {recording_id}')

  def insert_window(
      self,
      recording_id: int,
      offsets: list[float],
      embedding: np.ndarray | None = None,
      handle_duplicates: Literal[
          'allow', 'overwrite', 'skip', 'error'
      ] = 'error',
      **kwargs: Any,
  ) -> int:
    """Insert a window into the database."""

    if embedding is not None and embedding.shape[-1] != self._embedding_dim:
      raise ValueError(
          f'Incorrect embedding dimension. Expected {self._embedding_dim}, but'
          f' got {embedding.shape[-1]}.'
      )

    duplicate_id = self._handle_window_duplicates(
        recording_id, offsets, handle_duplicates
    )
    if duplicate_id is not None:
      return duplicate_id

    cursor = self._get_cursor()
    columns_str, placeholders_str, values = format_sql_insert_values(
        recording_id=recording_id,
        offsets=offsets,
        **kwargs,
    )
    try:
      cursor.execute(
          f"""
          INSERT INTO windows {columns_str}
          VALUES {placeholders_str}
          """,
          values,
      )
    except sqlite3.Error as e:
      if e.sqlite_errorname == 'SQLITE_CONSTRAINT_FOREIGNKEY':
        custom_msg = 'Check that the recording_id exists.'
      else:
        custom_msg = ''
      raise RuntimeError(
          f'Error inserting the window into the database. {custom_msg}'
      ) from e

    window_id = cursor.lastrowid
    if window_id is None:
      raise RuntimeError('Error inserting the window into the database.')
    if embedding is not None:
      if not self._ui_loaded:
        self.ui.load()
        self._ui_loaded = True
      self.ui.add(window_id, embedding.astype(self._embedding_dtype))
      self._ui_updated = True
    return window_id

  def insert_windows_batch(
      self,
      windows_batch: Sequence[dict[str, Any]],
      embeddings_batch: np.ndarray | None = None,
      handle_duplicates: Literal[
          'allow', 'overwrite', 'skip', 'error'
      ] = 'error',
  ) -> Sequence[int]:
    """Insert a batch of windows into the database."""

    if (
        embeddings_batch is not None
        and embeddings_batch.shape[-1] != self._embedding_dim
    ):
      raise ValueError(
          f'Incorrect embedding dimension. Expected {self._embedding_dim}, but'
          f' got {embeddings_batch.shape[-1]}.'
      )

    # Make sure that, unless we're in "allow" mode, there are no duplicates in
    # the batch itself.
    if handle_duplicates != 'allow':
      for i in range(len(windows_batch)):
        for j in range(i + 1, len(windows_batch)):
          if windows_batch[i]['recording_id'] == windows_batch[j][
              'recording_id'
          ] and np.allclose(
              windows_batch[i]['offsets'],
              windows_batch[j]['offsets'],
              rtol=0.0,
              atol=1e-6,
          ):
            raise RuntimeError(
                'Duplicates found in `windows_batch`, but this use case is not'
                ' supported unless `handle_duplicates` is set to "allow"'
                f' (handle_duplicates = "{handle_duplicates}").'
            )

    # Handle duplicates from the database.
    window_ids = [-1] * len(windows_batch)
    if handle_duplicates in ['overwrite', 'skip', 'error']:
      keep_idx = []
      remove_window_ids = set()

      # Go over all windows in the batch and check: (1) which ones need to be
      # removed, (2) which ones get a match and need to be skipped from
      # insertion, (3) which ones trigger an error, and (4) which ones still
      # need to be inserted.
      for idx, window_kwargs in enumerate(windows_batch):
        matches = self.get_all_windows(
            filter=config_dict.create(
                eq=dict(recording_id=window_kwargs['recording_id']),
                approx=dict(offsets=window_kwargs['offsets']),
            )
        )
        if matches:
          if handle_duplicates == 'overwrite':
            for match in matches:
              remove_window_ids.add(match.id)
            keep_idx.append(idx)
          elif handle_duplicates == 'skip':
            window_ids[idx] = matches[0].id
          elif handle_duplicates == 'error':
            raise RuntimeError(
                f'Duplicate window found (id = {matches[0].id}), but'
                ' `handle_duplicates` is set to "error".'
            )
        else:
          keep_idx.append(idx)

      # Now that no duplicates triggered an error, we can remove those that were
      # overwritten.
      for window_id in remove_window_ids:
        self.remove_window(window_id)

    else:
      keep_idx = list(range(len(windows_batch)))

    # Insert the windows that were not skipped and did not trigger an error.
    for idx in keep_idx:
      window_ids[idx] = self.insert_window(
          embedding=None,
          handle_duplicates='allow',
          **windows_batch[idx],
      )

    if embeddings_batch is not None:
      if not self._ui_loaded:
        self.ui.load()
        self._ui_loaded = True
      self.ui.add(
          np.array(window_ids)[keep_idx],
          embeddings_batch[keep_idx].astype(self._embedding_dtype),
      )
      self._ui_updated = True

    return window_ids

  def get_window(
      self,
      window_id: int,
      include_embedding: bool = False,
  ) -> datatypes.Window:
    """Get a window from the database."""

    window_id = int(window_id)

    cursor = self._get_cursor()
    cursor.execute(
        """
        SELECT *
        FROM windows
        WHERE id = ?
        """,
        (window_id,),
    )
    result = cursor.fetchone()
    if result is None:
      raise KeyError(f'Window id not found: {window_id}')

    columns = [col[0] for col in cursor.description]
    window = datatypes.Window(
        embedding=None,
        **dict(zip(columns, result)),
    )
    if include_embedding:
      window.embedding = self.get_embedding(window_id)
    return window

  def get_embedding(self, window_id: int) -> np.ndarray:
    """Get an embedding vector from the database."""

    window_id = int(window_id)

    found = self.ui.contains(window_id)
    if not isinstance(found, bool):
      raise RuntimeError(
          'Expected bool result from the USearch `contains()` method, but got'
          f' {type(found)}: {found}.'
      )
    if not found:
      raise KeyError(f'Embedding vector not found for window id: {window_id}')
    embedding = self.ui.get(window_id)
    if not isinstance(embedding, np.ndarray):
      raise RuntimeError(
          'Expected np.ndarray result from the USearch `get()` method, but got'
          f' {type(embedding)}: {embedding}.'
      )
    return embedding

  def get_embeddings_batch(
      self,
      window_ids: Sequence[int],
  ) -> np.ndarray:
    """Get a batch of embedding vectors from the database."""

    found = self.ui.contains(window_ids)
    if not isinstance(found, np.ndarray):
      raise RuntimeError(
          'Expected np.ndarray result from the USearch `contains()` method, but'
          f' got {type(found)}: {found}.'
      )
    if not np.all(found):
      raise KeyError(
          'Embedding vectors not found for window ids:'
          f' {itertools.compress(window_ids, ~found)}'
      )
    embeddings_batch = self.ui.get(window_ids)
    if not isinstance(embeddings_batch, np.ndarray):
      raise RuntimeError(
          'Expected np.ndarray result from the USearch `get()` method, but got'
          f' {type(embeddings_batch)}: {embeddings_batch}.'
      )
    return embeddings_batch

  def remove_window(self, window_id: int) -> None:
    """Remove a window from the database."""

    window_id = int(window_id)

    cursor = self._get_cursor()
    cursor.execute(
        """
        DELETE FROM windows
        WHERE id = ?
        """,
        (window_id,),
    )
    if cursor.rowcount == 0:
      raise KeyError(f'Window id not found: {window_id}')

    if not self._ui_loaded:
      self.ui.load()
      self._ui_loaded = True
    self.ui.remove(window_id)
    self._ui_updated = True

  def insert_annotation(
      self,
      recording_id: int,
      offsets: list[float],
      label: str,
      label_type: datatypes.LabelType,
      provenance: str,
      handle_duplicates: Literal[
          'allow', 'overwrite', 'skip', 'error'
      ] = 'error',
      **kwargs: Any,
  ) -> int:
    """Insert an annotation into the database."""

    duplicate_id = self._handle_annotation_duplicates(
        recording_id, offsets, label, label_type, provenance, handle_duplicates
    )
    if duplicate_id is not None:
      return duplicate_id

    cursor = self._get_cursor()
    columns_str, placeholders_str, values = format_sql_insert_values(
        recording_id=recording_id,
        offsets=offsets,
        label=label,
        label_type=label_type,
        provenance=provenance,
        **kwargs,
    )
    try:
      cursor.execute(
          f"""
          INSERT INTO annotations {columns_str}
          VALUES {placeholders_str}
          """,
          values,
      )
    except sqlite3.Error as e:
      if e.sqlite_errorname == 'SQLITE_CONSTRAINT_FOREIGNKEY':
        custom_msg = 'Check that the recording_id exists.'
      else:
        custom_msg = ''
      raise RuntimeError(
          f'Error inserting the annotation into the database. {custom_msg}'
      ) from e

    annotation_id = cursor.lastrowid
    if annotation_id is None:
      raise RuntimeError('Error inserting the annotation into the database.')
    return cursor.lastrowid

  def get_annotation(self, annotation_id: int) -> datatypes.Annotation:
    """Get an annotation from the database."""

    annotation_id = int(annotation_id)

    cursor = self._get_cursor()
    cursor.execute(
        """
        SELECT *
        FROM annotations
        WHERE id = ?
        """,
        (annotation_id,),
    )
    result = cursor.fetchone()
    if result is None:
      raise KeyError(f'Annotation id not found: {annotation_id}')

    columns = [col[0] for col in cursor.description]
    annotation = datatypes.Annotation(**dict(zip(columns, result)))
    annotation.label_type = datatypes.LabelType(annotation.label_type)
    return annotation

  def remove_annotation(self, annotation_id: int) -> None:
    """Remove an annotation from the database."""

    annotation_id = int(annotation_id)

    cursor = self._get_cursor()
    cursor.execute(
        """
        DELETE FROM annotations
        WHERE id = ?
        """,
        (annotation_id,),
    )
    if cursor.rowcount == 0:
      raise KeyError(f'Annotation id not found: {annotation_id}')

  def count_embeddings(self) -> int:
    """Get the number of embeddings in the database."""
    return self.ui.size

  def match_window_ids(
      self,
      deployments_filter: config_dict.ConfigDict | None = None,
      recordings_filter: config_dict.ConfigDict | None = None,
      windows_filter: config_dict.ConfigDict | None = None,
      annotations_filter: config_dict.ConfigDict | None = None,
      limit: int | None = None,
  ) -> Sequence[int]:
    """Get matching window IDs from the database based on given filters."""

    # TODO(stefanistrate): Is it more efficient for large databases to run
    # multiple SELECT queries (in order: by deployment, by recording, by window,
    # by annotation) instead of running one giant JOIN?

    cursor = self._get_cursor()

    # Pick which tables need to be queried.
    query_tables = {'windows'}
    if annotations_filter:
      query_tables |= {'annotations'}
    if recordings_filter:
      query_tables |= {'recordings'}
    if deployments_filter:
      query_tables |= {'recordings', 'deployments'}

    # Build the `SELECT ... FROM ... [JOIN ...]` part of the SQL query.
    if 'annotations' in query_tables:
      select_clause = 'SELECT DISTINCT windows.id'
    else:
      select_clause = 'SELECT windows.id'
    from_clause = 'FROM windows'
    if 'annotations' in query_tables:
      from_clause += (
          ' JOIN annotations ON windows.recording_id = annotations.recording_id'
          ' AND APPROX_FLOAT_LIST(windows.offsets, annotations.offsets) = TRUE'
      )
    if 'recordings' in query_tables:
      from_clause += ' JOIN recordings ON windows.recording_id = recordings.id'
    if 'deployments' in query_tables:
      from_clause += (
          ' JOIN deployments ON recordings.deployment_id = deployments.id'
      )

    # Build the `WHERE ...` part of the SQL query.
    conditions, values = tuple(
        zip(*[
            format_sql_where_conditions(
                annotations_filter, table_prefix='annotations'
            ),
            format_sql_where_conditions(windows_filter, table_prefix='windows'),
            format_sql_where_conditions(
                recordings_filter, table_prefix='recordings'
            ),
            format_sql_where_conditions(
                deployments_filter, table_prefix='deployments'
            ),
        ])
    )
    conditions_str = ' AND '.join(c for c in conditions if c)
    values = list(itertools.chain.from_iterable(values))
    where_clause = f'WHERE {conditions_str}' if conditions_str else ''

    # Build the `LIMIT ...` part of the SQL query.
    if limit is None:
      limit_clause = ''
    else:
      limit_clause = f'LIMIT {limit}'

    # Execute the SQL query and return the window IDs.
    cursor.execute(
        f"""
        {select_clause}
        {from_clause}
        {where_clause}
        {limit_clause}
        """,
        values,
    )
    return [result[0] for result in cursor.fetchall()]

  def get_all_projects(self) -> Sequence[str]:
    """Get all distinct projects from the database."""

    cursor = self._get_cursor()
    cursor.execute("""
        SELECT DISTINCT project
        FROM deployments
        ORDER BY project
        """)
    return [result[0] for result in cursor.fetchall()]

  def get_all_deployments(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Deployment]:
    """Get all deployments from the database."""

    cursor = self._get_cursor()
    conditions_str, values = format_sql_where_conditions(filter)
    where_clause = f'WHERE {conditions_str}' if conditions_str else ''
    cursor.execute(
        f"""
        SELECT *
        FROM deployments
        {where_clause}
        """,
        values,
    )

    columns = [col[0] for col in cursor.description]
    deployments = [
        datatypes.Deployment(**dict(zip(columns, result)))
        for result in cursor.fetchall()
    ]
    return deployments

  def get_all_recordings(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Recording]:
    """Get all recordings from the database."""

    cursor = self._get_cursor()
    conditions_str, values = format_sql_where_conditions(filter)
    where_clause = f'WHERE {conditions_str}' if conditions_str else ''
    cursor.execute(
        f"""
        SELECT *
        FROM recordings
        {where_clause}
        """,
        values,
    )

    recordings = []
    columns = [col[0] for col in cursor.description]
    for result in cursor.fetchall():
      recording = datatypes.Recording(**dict(zip(columns, result)))
      if recording.datetime is not None:
        recording.datetime = dt.datetime.fromisoformat(recording.datetime)
      recordings.append(recording)
    return recordings

  def get_all_windows(
      self,
      include_embedding: bool = False,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Window]:
    """Get all windows from the database."""

    cursor = self._get_cursor()
    conditions_str, values = format_sql_where_conditions(filter)
    where_clause = f'WHERE {conditions_str}' if conditions_str else ''
    cursor.execute(
        f"""
        SELECT *
        FROM windows
        {where_clause}
        """,
        values,
    )

    windows = []
    columns = [col[0] for col in cursor.description]
    for result in cursor.fetchall():
      window = datatypes.Window(
          embedding=None,
          **dict(zip(columns, result)),
      )
      if include_embedding:
        window.embedding = self.get_embedding(window.id)
      windows.append(window)
    return windows

  def get_all_annotations(
      self,
      filter: config_dict.ConfigDict | None = None,  # pylint: disable=redefined-builtin
  ) -> Sequence[datatypes.Annotation]:
    """Get all annotations from the database."""

    cursor = self._get_cursor()
    conditions_str, values = format_sql_where_conditions(filter)
    where_clause = f'WHERE {conditions_str}' if conditions_str else ''
    cursor.execute(
        f"""
        SELECT *
        FROM annotations
        {where_clause}
        """,
        values,
    )

    annotations = []
    columns = [col[0] for col in cursor.description]
    for result in cursor.fetchall():
      annotation = datatypes.Annotation(**dict(zip(columns, result)))
      annotation.label_type = datatypes.LabelType(annotation.label_type)
      annotations.append(annotation)
    return annotations

  def get_all_labels(
      self,
      label_type: datatypes.LabelType | None = None,
  ) -> Sequence[str]:
    """Get all distinct labels from the database."""

    cursor = self._get_cursor()
    if label_type is None:
      where_clause = ''
      values = tuple()
    else:
      filter_dict = config_dict.create(eq=dict(label_type=label_type))
      conditions_str, values = format_sql_where_conditions(filter_dict)
      where_clause = f'WHERE {conditions_str}' if conditions_str else ''
    cursor.execute(
        f"""
        SELECT DISTINCT label
        FROM annotations
        {where_clause}
        ORDER BY label
        """,
        values,
    )
    return [result[0] for result in cursor.fetchall()]

  def count_each_label(
      self,
      label_type: datatypes.LabelType | None = None,
  ) -> collections.Counter[str]:
    """Count each label in the database, ignoring provenance."""

    cursor = self._get_cursor()
    if label_type is None:
      where_clause = ''
      values = tuple()
    else:
      filter_dict = config_dict.create(eq=dict(label_type=label_type))
      conditions_str, values = format_sql_where_conditions(filter_dict)
      where_clause = f'WHERE {conditions_str}' if conditions_str else ''

    # Subselect with DISTINCT is needed to avoid double-counting the same label
    # on the same recording offsets because of different provenances.
    cursor.execute(
        f"""
        SELECT label, COUNT(*)
        FROM (
            SELECT DISTINCT recording_id, offsets, label, label_type
            FROM annotations
            {where_clause}
        )
        GROUP BY label
        ORDER BY label
        """,
        values,
    )
    return collections.Counter(
        {result[0]: result[1] for result in cursor.fetchall()}
    )

  def get_embedding_dim(self) -> int:
    """Get the embedding dimension."""
    return self._embedding_dim

  def get_embedding_dtype(self) -> type[Any]:
    """Get the embedding data type."""
    return self._embedding_dtype

  def search(
      self,
      query_embedding: np.ndarray,
      search_list_size: int,
      approximate: bool = True,
      target_score: float | None = None,
      score_fn_name: str = 'dot',
      **kwargs: Any,
  ) -> search_results.TopKSearchResults:
    """Search for neighbors of the query embedding.

    Args:
      query_embedding: The query embedding vector.
      search_list_size: The number of results to return.
      approximate: Whether to use approximate search.
      target_score: If set, search for examples near this score.
      score_fn_name: The name of the score function to use. Likely ignored if
        `approximate` is True, as the underlying indexed score function will be
        used.
      **kwargs: Additional keyword arguments to pass to the search function.

    Returns:
      A TopKSearchResults object containing the search results.
    """
    if not approximate and target_score is not None:
      return super().search(
          query_embedding=query_embedding,
          search_list_size=search_list_size,
          approximate=approximate,
          target_score=target_score,
          score_fn_name=score_fn_name,
          **kwargs,
      )

    if target_score is not None:
      raise ValueError(
          'Approximate search does not support target_score sampling.'
      )

    ann_matches = self.ui.search(
        query_embedding, count=search_list_size, exact=not approximate
    )
    results = search_results.TopKSearchResults(top_k=search_list_size)
    for k, d in zip(ann_matches.keys, ann_matches.distances):
      # Usearch with IP returns 1 - inner_product.
      # We want higher is better, so 1 - d.
      results.update(search_results.SearchResult(k, 1.0 - d))
    score_fn = score_functions.get_score_fn(score_fn_name)
    results = brutalism.rerank(query_embedding, results, self, score_fn)
    return results



# coding=utf-8
# Copyright 2026 The Perch Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helpers for loading specific annotation CSVs."""

import json
import os

from etils import epath
import pandas as pd
from perch_hoplite.taxonomy import annotations
from perch_hoplite.taxonomy import namespace_db

_DEPRECATED2NEW = {
    'mallar': 'mallar3',
    'rufant1': 'rufant7',
}


def load_caples_annotations(annotations_path: epath.Path) -> pd.DataFrame:
  """Loads the dataframe of all caples annotations from annotation CSV.

  Args:
    annotations_path: Filepath for the annotations CSV.

  Returns:
    DataFrame of annotations.
  """
  filename_fn = lambda _, row: row['fid'].strip()
  start_time_fn = lambda row: float(row['start_time_s'])
  end_time_fn = lambda row: float(row['end_time_s'])
  # Get rid of the one bad label in the dataset...
  filter_fn = lambda row: 'comros' in row['ebird_codes']
  class_fn = lambda row: row['ebird_codes'].split(' ')
  annos = annotations.read_dataset_annotations_csvs(
      [annotations_path],
      filename_fn=filename_fn,
      namespace='ebird2021',
      class_fn=class_fn,
      start_time_fn=start_time_fn,
      end_time_fn=end_time_fn,
      filter_fn=filter_fn,
  )
  segments = annotations.annotations_to_dataframe(annos)
  return segments


def load_cornell_annotations(
    annotations_path: epath.Path, file_id_prefix: str = ''
) -> pd.DataFrame:
  """Load the annotations from a Cornell Zenodo dataset."""
  start_time_fn = lambda row: float(row['Start Time (s)'])
  end_time_fn = lambda row: float(row['End Time (s)'])
  filter_fn = lambda row: False
  class_fn = lambda row: [  # pylint: disable=g-long-lambda
      row['Species eBird Code'].strip().replace('????', 'unknown')
  ]

  filename_fn = lambda filepath, row: file_id_prefix + row['Filename'].strip()
  annos = annotations.read_dataset_annotations_csvs(
      [annotations_path],
      filename_fn=filename_fn,
      namespace='ebird2021',
      class_fn=class_fn,
      start_time_fn=start_time_fn,
      end_time_fn=end_time_fn,
      filter_fn=filter_fn,
  )
  segments = annotations.annotations_to_dataframe(annos)
  return segments


def load_powdermill_annotations(annotations_path: epath.Path) -> pd.DataFrame:
  """Load annotations from https://zenodo.org/records/4656848."""
  start_time_fn = lambda row: float(row['Begin Time (s)'])
  end_time_fn = lambda row: float(row['End Time (s)'])
  filter_fn = lambda row: False

  # Convert dataset labels to ebird2021.
  db = namespace_db.load_db()
  ebird_mapping = db.mappings['ibp2019_to_ebird2021']
  ebird_mapping_dict = ebird_mapping.mapped_pairs
  class_fn = lambda row: [  # pylint: disable=g-long-lambda
      ebird_mapping_dict.get(row['Species'].strip(), row['Species'].strip())
  ]

  annotation_filepaths = [annotations_path]
  filename_fn = lambda filepath, row: row['Filename'].strip()
  annos = annotations.read_dataset_annotations_csvs(
      annotation_filepaths,
      filename_fn=filename_fn,
      namespace=ebird_mapping.target_namespace,
      class_fn=class_fn,
      start_time_fn=start_time_fn,
      end_time_fn=end_time_fn,
      filter_fn=filter_fn,
  )
  segments = annotations.annotations_to_dataframe(annos)
  return segments


def load_weldy_annotations(annotations_path: epath.Path) -> pd.DataFrame:
  """Loads annotations from https://zenodo.org/records/10252138."""
  filename_fn = lambda _, row: 'annotated_recordings/' + row['file'].strip()
  start_time_fn = lambda row: float(row['start'])
  end_time_fn = lambda row: float(row['end'])
  filter_fn = lambda row: False
  class_fn = lambda row: (  # pylint: disable=g-long-lambda
      row['label']
      .replace('unk', 'unknown')
      .replace('impossible', 'unknown')
      .replace('unknown_chip', 'unknown')
      .split(' ')
  )
  annos = annotations.read_dataset_annotations_csvs(
      [epath.Path(annotations_path)],
      filename_fn=filename_fn,
      namespace='weldy_calltype',
      class_fn=class_fn,
      start_time_fn=start_time_fn,
      end_time_fn=end_time_fn,
      filter_fn=filter_fn,
  )
  segments = annotations.annotations_to_dataframe(annos)
  return segments


def load_anuraset_annotations(
    annotations_path: epath.Path, prefix: str = 'raw_data'
) -> pd.DataFrame:
  """Loads raw audio annotations from https://zenodo.org/records/8342596."""
  filename_fn = lambda _, row: os.path.join(  # pylint: disable=g-long-lambda
      prefix, row['filename'].split('_')[0], row['filename'].strip()
  )
  start_time_fn = lambda row: float(row['start_time_s'])
  end_time_fn = lambda row: float(row['end_time_s'])
  # There are a few SPECIES_LALSE labels which according to the authors should
  # be ignored.
  filter_fn = lambda row: '_LALSE' in row['label']
  class_fn = lambda row: row['label'].split(' ')
  annos = annotations.read_dataset_annotations_csvs(
      [epath.Path(annotations_path)],
      filename_fn=filename_fn,
      namespace='anuraset',
      class_fn=class_fn,
      start_time_fn=start_time_fn,
      end_time_fn=end_time_fn,
      filter_fn=filter_fn,
  )
  segments = annotations.annotations_to_dataframe(annos)
  return segments


def load_reef_annotations(annotations_path: epath.Path) -> pd.DataFrame:
  """Loads a dataframe of all annotations from the ReefSet JSON file.

  Args:
    annotations_path: path to dataset_v*.json.

  Returns:
    DataFrame of metadata parsed from the datasets
    Reef specific stuff:
    - All clips are 1.88sec long, so we fix all start and end times accordingly
    - We only take entries for which the dataset_type is sound_event_dataset, as
    other entries are only soundscape (habitat level) labels or just unlabeled
    completely
    - In future, should this add a header to the df that species the region
    somehow? Allowing selection by regional datasets
  """
  # Read the JSON file
  with annotations_path.open() as f:
    data = json.load(f)
  # Prepare a list of dictionaries for creating a DataFrame
  rows = []
  for entry in data:
    # Include only entries with "dataset_type": "sound_event_dataset"
    if entry.get('dataset_type') == 'sound_event_dataset':
      label = entry.get('label', '')
      # to use region.label format use:
      # label = f"{entry.get('region', '')}.{entry.get('label', '')}"
      row = {
          'filename': entry.get('file_name', ''),
          'start_time_s': 0.0,
          'end_time_s': 1.88,
          'namespace': 'reefs',
          'label': [label],
      }
      rows.append(row)
  # Create a DataFrame
  segments = pd.DataFrame(rows)
  return segments
